from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from importlib.resources import path
from typing import ContextManager


@dataclass
class ResourcePath:
    base_module: str
    file_name: str

    @contextmanager
    def use(self) -> ContextManager[Path]:
        with path(self.base_module, Path(self.file_name)) as use_path:
            yield use_path



import hashlib
from pathlib import Path
from typing import List, Dict, Any, Optional, Tuple

from pkm.api.dependencies.dependency import Dependency
from pkm.api.environments.environment import Environment
from pkm.api.packages.package import Package, PackageDescriptor
from pkm.api.packages.package_metadata import PackageMetadata
from pkm.api.packages.standard_package import AbstractPackage, PackageArtifact
from pkm.api.repositories.repository import Authentication, AbstractRepository
from pkm.api.repositories.repository import RepositoryPublisher
from pkm.api.versions.version import Version
from pkm.api.versions.version_specifiers import VersionSpecifier
from pkm.utils.commons import NoSuchElementException
from pkm.utils.http.cache_directive import CacheDirective
from pkm.utils.http.http_client import HttpClient, HttpException
from pkm.utils.http.mfd_payload import FormField, MultipartFormDataPayload
from pkm.utils.io_streams import chunks
from pkm.utils.properties import cached_property


class PyPiRepository(AbstractRepository):

    def __init__(self, http: HttpClient):
        super().__init__('pypi')
        self._http = http

    @cached_property
    def publisher(self) -> Optional["RepositoryPublisher"]:
        return PyPiPublisher(self._http)

    def _do_match(self, dependency: Dependency) -> List[Package]:
        # monitor.on_dependency_match(dependency)
        try:
            json: Dict[str, Any] = self._http \
                .fetch_resource(f'https://pypi.org/pypi/{dependency.package_name}/json',
                                cache=CacheDirective.ask_for_update(),
                                resource_name=f"matching packages for {dependency}") \
                .read_data_as_json()
        except HttpException as e:
            raise NoSuchElementException(
                f"package: '{dependency.package_name}' could not be retrieved from repository: '{self.name}'") from e

        package_info: Dict[str, Any] = {k.replace('_', '-').title(): v for k, v in json['info'].items()}

        packages: List[PypiPackage] = []
        releases: Dict[str, Any] = json['releases']
        for version_str, release in releases.items():

            relevant_artifacts = []
            for a in release:
                if not a.get('yanked') and a.get("packagetype") in ("sdist", "bdist_wheel"):
                    if sa := _create_artifact_from_pypi_release(a):
                        relevant_artifacts.append(sa)

            if relevant_artifacts:
                version = Version.parse(version_str)
                if dependency.version_spec.allows_version(version):
                    packages.append(PypiPackage(
                        PackageDescriptor(dependency.package_name, version),
                        relevant_artifacts, self, PackageMetadata(path=None, data=package_info)
                    ))

        return packages


# noinspection PyProtectedMember
class PypiPackage(AbstractPackage):

    def __init__(self, descriptor: PackageDescriptor, artifacts: List[PackageArtifact], repo: PyPiRepository,
                 metadata: PackageMetadata):

        super().__init__(descriptor, artifacts, metadata)
        self._repo = repo

    def _retrieve_artifact(self, artifact: PackageArtifact) -> Path:
        url = artifact.other_info.get('url')
        if not url:
            raise KeyError(f'could not find url in given artifact info: {artifact}')

        resource = self._repo._http.fetch_resource(url, resource_name=f"{self.name} {self.version}")
        if not resource:
            raise FileNotFoundError(f'cannot find requested artifact: {artifact.file_name}')

        return resource.data

    def _all_dependencies(self, environment: "Environment") -> List["Dependency"]:
        json: Dict[str, Any] = self._repo._http \
            .fetch_resource(f'https://pypi.org/pypi/{self.name}/{self.version}/json',
                            resource_name=f"metadata for {self.name} {self.version}") \
            .read_data_as_json()

        requires_dist = json['info'].get('requires_dist')
        if requires_dist is None:
            return super(PypiPackage, self)._all_dependencies(environment)

        return [Dependency.parse(dstr) for dstr in requires_dist]


def _create_artifact_from_pypi_release(release: Dict[str, Any]) -> Optional[PackageArtifact]:
    requires_python = release.get('requires_python')

    file_name: str = release.get('filename')
    if not file_name:
        return None

    return PackageArtifact(
        file_name, VersionSpecifier.parse(requires_python) if requires_python else None, release
    )


# https://warehouse.pypa.io/api-reference/legacy.html
class PyPiPublisher(RepositoryPublisher):

    def __init__(self, http: HttpClient):
        super().__init__('pypi')
        self._http = http

    def publish(self, auth: "Authentication", package_meta: PackageMetadata, distribution: Path):
        print(f"uploading distribution: {distribution}")

        data = {k.replace('-', '_').lower(): v for k, v in package_meta.items()}
        file_type = 'bdist_wheel' if distribution.suffix == '.whl' else 'sdist'
        py_version = distribution.name.split("-")[2] if distribution.suffix == '.whl' else 'source'

        md5, sha256, blake2 = hashlib.md5(), hashlib.sha256(), hashlib.blake2b(digest_size=256 // 8)
        with distribution.open('rb') as d_fd:
            for chunk in chunks(d_fd):
                md5.update(chunk)
                sha256.update(chunk)
                blake2.update(chunk)

        data.update({
            'filetype': file_type,
            'pyversion': py_version,
            'md5_digest': md5.hexdigest(),
            'sha256_digest': sha256.hexdigest(),
            'blake2_256_digest': blake2.hexdigest(),
            ':action': 'file_upload',
            'protocol_version': '1'
        })

        fields: List[FormField] = []
        for k, v in data.items():
            if isinstance(v, (Tuple, List)):
                for iv in v:
                    fields.append(FormField(k, iv))
            else:
                fields.append(FormField(k, v))

        with distribution.open('rb') as d_fd:
            fields.append(FormField('content', d_fd, filename=distribution.name)
                          .set_content_type("application/octet-stream"))

            payload = MultipartFormDataPayload(fields=fields)
            headers = dict([
                auth.as_basic_auth_header(),
                ("Content-Type", payload.content_type()),
            ])

            # for tests we can use: "https://test.pypi.org/legacy/"
            # with self._http.post("https://test.pypi.org/legacy/", payload, headers=headers,
            with self._http.post("https://upload.pypi.org/legacy/", payload, headers=headers,
                                 max_redirects=0) as response:
                if response.status != 200:
                    content = response.read()
                    raise HttpException(f"publish failed, server responded with {response.status}, {content}")



from __future__ import annotations

import warnings
from pathlib import Path
from typing import List, Dict, Optional, Tuple, Iterable

from pkm.api.dependencies.dependency import Dependency
from pkm.api.environments.environment import Environment
from pkm.api.environments.environments_zoo import EnvironmentsZoo
from pkm.api.packages.package import Package, PackageDescriptor
from pkm.api.projects.project import Project
from pkm.api.projects.project_group import ProjectGroup
from pkm.api.repositories.repositories_configuration import RepositoryInstanceConfig, RepositoriesConfiguration, \
    RepositoriesConfigInheritanceMode
from pkm.api.repositories.repository import Repository, RepositoryBuilder, AbstractRepository
from pkm.repositories.file_repository import FileRepository
from pkm.repositories.shared_pacakges_repo import SharedPackagesRepository
from pkm.resolution.packages_lock import LockPrioritizingRepository
from pkm.utils.commons import NoSuchElementException
from pkm.utils.dicts import put_if_absent
from pkm.utils.http.http_client import HttpClient

REPOSITORIES_ENTRYPOINT_GROUP = "pkm-repositories"
REPOSITORIES_CONFIGURATION_PATH = "etc/pkm/repositories.toml"


class RepositoryLoader:
    def __init__(self, main_cfg: Path, http: HttpClient, workspace: Path):

        from pkm.api.environments.environment import Environment
        from pkm.repositories.simple_repository import SimpleRepositoryBuilder
        from pkm.repositories.git_repository import GitRepository
        from pkm.repositories.pypi_repository import PyPiRepository
        from pkm.repositories.local_packages_repository import LocalPackagesRepositoryBuilder
        from pkm.repositories.url_repository import UrlRepository

        # base repositories
        self.pypi = PyPiRepository(http)
        self._cached_instances: Dict[RepositoryInstanceConfig, Repository] = {}  # noqa

        self._url_repos = {
            r.name: r for r in (
                GitRepository(workspace / 'git'),
                UrlRepository(),
                FileRepository()
            )
        }
        self._url_repos['http'] = self._url_repos['https'] = self._url_repos['url']

        # common builders
        self._builders = {
            b.name: b for b in (
                SimpleRepositoryBuilder(http),
                LocalPackagesRepositoryBuilder(),
            )
        }

        # builders from entrypoints
        for epoint in Environment.current().entrypoints[REPOSITORIES_ENTRYPOINT_GROUP]:
            try:
                builder: RepositoryBuilder = epoint.ref.import_object()()
                if not isinstance(builder, RepositoryBuilder):
                    raise ValueError("repositories entrypoint did not point to a repository builder class")
            except Exception:  # noqa
                import traceback
                warnings.warn(f"malformed repository entrypoint: {epoint}")
                traceback.print_exc()

        self._main = self._compose('main', main_cfg, self.pypi)
        self.workspace = workspace

    @property
    def main(self) -> Repository:
        return self._with_url_support(self._main)

    def _with_url_support(self, repo: Repository) -> Repository:
        return _UrlSupportedRepository(self._url_repos, repo)

    def load_for_env_zoo(self, zoo: EnvironmentsZoo) -> Repository:
        config = zoo.path / REPOSITORIES_CONFIGURATION_PATH
        repo = self._main
        if config.exists():
            repo = self._compose("zoo-configured-repository", config, repo)

        if zoo.config.package_sharing.enabled:
            repo = SharedPackagesRepository(zoo.path / ".zoo/shared", repo)

        return self._with_url_support(repo)

    def load_for_env(self, env: Environment) -> Repository:
        repo = self._main
        if zoo := env.zoo:
            repo = zoo.attached_repository

        config = env.path / REPOSITORIES_CONFIGURATION_PATH
        if config.exists():
            repo = self._compose("env-configured-repository", config, repo)

        return self._with_url_support(repo)

    def load_for_project(self, project: Project) -> Repository:
        repo = project.attached_environment.attached_repository

        if group := project.group:
            repo = self._load_for_project_group(group, repo)
        else:
            repo = _ProjectsRepository.create('project-repository', [project], repo)

        config = project.path / REPOSITORIES_CONFIGURATION_PATH
        if config.exists():
            repo = self._compose("project-configured-repository", config, repo)

        repo = LockPrioritizingRepository(
            "lock-prioritizing-repository", repo, project.lock,
            project.attached_environment)

        return self._with_url_support(repo)

    def load_for_project_group(self, group: ProjectGroup) -> Repository:
        return self._with_url_support(self._load_for_project_group(group, None))

    def _load_for_project_group(self, group: ProjectGroup, base_repo: Optional[Repository] = None) -> Repository:
        repo = base_repo or self._main

        config = group.path / REPOSITORIES_CONFIGURATION_PATH
        if config.exists():
            repo = self._compose("group-configured-repository", config, repo)

        repo = _ProjectsRepository.create("group-projects-repository", group.project_children_recursive, repo)
        return repo

    def _compose(self, name: str, config_path: Path, next_in_context: Repository) -> Repository:
        package_search_list = []
        package_associated_repo = {}

        config = RepositoriesConfiguration.load(config_path)

        if not (new_repos := config.repositories):
            if config.inheritance_mode == RepositoriesConfigInheritanceMode.INHERIT_MAIN:
                return self.main
            return next_in_context

        for definition in new_repos:
            instance = self.build(definition)

            if definition.packages:
                for package in definition.packages:
                    if package == '*':
                        package_search_list.append(instance)
                    else:
                        put_if_absent(package_associated_repo, package, instance)
            else:
                package_search_list.append(instance)

        if config.inheritance_mode == RepositoriesConfigInheritanceMode.INHERIT_CONTEXT:
            package_search_list.append(next_in_context)
        elif config.inheritance_mode == RepositoriesConfigInheritanceMode.INHERIT_MAIN:
            package_search_list.append(self.main)

        return _CompositeRepository(name, package_search_list, package_associated_repo)

    def build(self, config: RepositoryInstanceConfig) -> Repository:
        if not (cached := self._cached_instances.get(config)):
            if not (builder := self._builders.get(config.type)):
                raise KeyError(f"unknown repository type: {config.type}")
            cached = builder.build(config.name, config.packages, **config.args)
            self._cached_instances[config] = cached

        return cached


class _UrlSupportedRepository(AbstractRepository):
    def __init__(self, url_handlers: Dict[str, Repository], repo: Repository):
        super().__init__("url-supported-repo")
        self._url_handlers = url_handlers
        self._base_repo = repo

    def _do_match(self, dependency: Dependency) -> List[Package]:
        if url := dependency.version_spec.specific_url():

            if protocol := url.protocol:
                if repo := self._url_handlers.get(url.protocol):
                    return repo.match(dependency, False)
                raise NoSuchElementException(f"could not find repository to handle url with protocol: {protocol}")
            return self._url_handlers['url'].match(dependency, False)

        return self._base_repo.match(dependency, False)

    def _sort_by_priority(self, dependency: Dependency, packages: List[Package]) -> List[Package]:
        return packages


class _CompositeRepository(AbstractRepository):
    def __init__(
            self, name: str, package_search_list: List[Repository],
            package_associated_repos: Dict[str, Repository]):
        super().__init__(name)

        self._package_search_list = package_search_list
        self._package_associated_repos = package_associated_repos

    def _do_match(self, dependency: Dependency) -> List[Package]:
        if repo := self._package_associated_repos.get(dependency.package_name):
            return repo.match(dependency, False)

        for repo in self._package_search_list:
            if result := repo.match(dependency, False):
                return result

        return []

    def _sort_by_priority(self, dependency: Dependency, packages: List[Package]) -> List[Package]:
        return packages


class _ProjectsRepository(AbstractRepository):
    def __init__(self, name: str, projects: Dict[str, Tuple[PackageDescriptor, Path]], base_repo: Repository):
        super().__init__(name)
        self._packages = projects
        self._base_repo = base_repo

    def _do_match(self, dependency: Dependency) -> List[Package]:
        if (package_and_path := self._packages.get(dependency.package_name)) and \
                dependency.version_spec.allows_version(package_and_path[0].version):
            return [Project.load(package_and_path[1])]
        return self._base_repo.match(dependency, False)

    def _sort_by_priority(self, dependency: Dependency, packages: List[Package]) -> List[Package]:
        return packages

    @classmethod
    def create(cls, name: str, projects: Iterable[Project], base: Repository) -> _ProjectsRepository:
        return _ProjectsRepository(name, {p.name: (p.descriptor, p.path) for p in projects}, base)



from __future__ import annotations

import json
import os
import re
import subprocess
from dataclasses import dataclass
from pathlib import Path
from struct import Struct
from typing import Iterator, Tuple, List, Optional, Dict, IO, Callable

from pkm.api.versions.version import StandardVersion
from pkm.config.configuration import Configuration
from pkm.utils.files import temp_dir
from pkm.utils.properties import cached_property
from pkm.utils.sequences import index_of_or_none
from pkm.utils.types import SupportsLessThan, SupportsLessThanEq

_INTROSPECTION_CODE = """
import sys
import sysconfig
import json
import importlib.machinery
import platform
import os
import site
import struct

if hasattr(sys, 'implementation'):
    sys_implementation = {'name': sys.implementation.name, 'version': sys.implementation.version}
else:
    sys_implementation = {'name': platform.python_implementation().lower()}

result = {
    'sysconfig': {
        'paths': sysconfig.get_paths(),
        'paths_by_scheme': {s:sysconfig.get_paths(s) for s in sysconfig.get_scheme_names()},
        'vars': sysconfig.get_config_vars(),
        'platform': sysconfig.get_platform(),
        'is_python_build': sysconfig.is_python_build(True)
    },

    'sys': {
        'implementation': sys_implementation,
         'version_info': sys.version_info,
         'has_totalrefcount': hasattr(sys,'gettotalrefcount'),
         'maxunicode': sys.maxunicode,
         'maxsize': sys.maxsize,
         'executable': sys.executable,
         'platform': sys.platform
    },

    'importlib': {
        'machinery': {
            'EXTENSION_SUFFIXES': importlib.machinery.EXTENSION_SUFFIXES
        }
    },

    'platform': {
        'system': platform.system(),
        'mac_ver': platform.mac_ver(),
        'machine': platform.machine(),
        'python_implementation': platform.python_implementation(),
        'release': platform.release(),
        'version': platform.version(),
        'python_version_tuple': platform.python_version_tuple(),
        'python_version': platform.python_version()
    },

    'os': {
        'conf': {cname: os.confstr(cname) for cname in getattr(os,'confstr_names',[])},
        'name': os.name,
        'bits': struct.calcsize('P') * 8 
    },
    
    'site': {
        'packages': site.getsitepackages()
    }
}

print(json.dumps(result))
"""

_USER_SCHEME_BY_PLATFORM = {
    "Linux": "posix_user",
    "Windows": "nt_user",
    "Darwin": "osx_framework_user"
}


# noinspection PyRedundantParentheses
class EnvironmentIntrospection(Configuration):

    @property
    def paths(self) -> Dict[str, str]:
        return self._data['sysconfig']['paths']

    def is_windows_env(self) -> bool:
        return self['sys']['platform'] == 'win32'

    def user_paths(self) -> Optional[Dict[str, str]]:
        scheme = _USER_SCHEME_BY_PLATFORM.get(self._data["platform"]["system"])
        if not scheme:
            return None

        return self._data["sysconfig"]["paths_by_scheme"].get(scheme)

    def compute_markers(self) -> Dict[str, str]:
        """
        :return: pep508 environment markers
        """

        def _compute_implementation_version():
            if info := self._data['sys']['implementation'].get('version'):
                version = '.'.join(str(it) for it in info[:3])
                kind = info[3]
                if kind != 'final':
                    version += kind[0] + str(info[4])

                return version
            return "0"

        return {
            "os_name": self._data['os']['name'],
            "sys_platform": self._data['sys']['platform'],
            "platform_machine": self._data['platform']['machine'],
            "platform_python_implementation": self._data['platform']['python_implementation'],
            "platform_release": self._data['platform']['release'],
            "platform_system": self._data['platform']['system'],
            "platform_version": self._data['platform']['version'],
            "python_version": '.'.join(self._data['platform']['python_version_tuple'][:2]),
            "python_full_version": self._data['platform']['python_version'],
            "implementation_name": self._data['sys']['implementation']['name'],
            "implementation_version": _compute_implementation_version()
        }

    def compatibility_score(self, tag: str) -> Optional[SupportsLessThanEq]:
        """
        compute the compatibility score for the given pep425 compatibility tag
        :param tag: the pep425 compatibility tag
        :return: an opaque score object that support __le__ and __eq__ operations (read: comparable)
                 which can be treated as a score (read: higher is better)
        """
        try:
            max_score = None
            interpreters, abis, platforms = tag.split("-")
            for interpreter in interpreters.split("."):
                for abi in abis.split("."):
                    for platform_ in platforms.split("."):
                        score = self._compatibility_tag_scorer(interpreter, abi, platform_)

                        if score is not None and (max_score is None or max_score < score):
                            max_score = score
            return max_score
        except:  # noqa
            import traceback
            traceback.print_exc()
            return None

    @property
    def interpreter_name(self) -> str:
        """
        :return: the name of the interpreter as given in `sys.implementation.name`
        """
        iname = self._data['sys']['implementation']['name']
        return INTERPRETER_SHORT_NAMES.get(iname, iname)

    @property
    def interpreter_path(self) -> str:
        return self._data['sys']['executable']

    @property
    def interpreter_version(self) -> List[int]:
        return self._data['sys']['version_info']

    def _compute_generic_abis(self) -> Iterator[str]:
        if abi := self._data['sysconfig']['vars'].get("SOABI"):
            yield _normalize_string(abi)

    @property
    def is_32bit_interpreter(self):
        return self._data['sys']['maxsize'] <= 2 ** 32

    @cached_property
    def _glibc_version(self) -> Optional[StandardVersion]:
        # this is correct that the pkm's os and the actual os that will execute this
        # script should be the same but this is a future proofing for venvs in dockers
        # which is a use case pkm might end up supporting
        version_string = self._data['os']['conf'].get('CS_GNU_LIBC_VERSION')
        if not version_string:
            return None

        try:
            # os.confstr("CS_GNU_LIBC_VERSION") returns a string like "glibc 2.17".
            return StandardVersion.parse(version_string.split()[1])
        except ValueError:
            return None

    def _compute_cpython_abis(self, py_version: Tuple[int, ...]) -> List[str]:
        abis = []
        version = ''.join(str(v) for v in py_version)
        debug = pymalloc = ucs4 = ""
        with_debug = self._data['sysconfig']['vars'].get("Py_DEBUG")
        has_refcount = self._data['sys']['has_totalrefcount']
        # Windows doesn't set Py_DEBUG, so checking for support of debug-compiled
        # extension modules is the best option.
        # https://github.com/pypa/pip/issues/3383#issuecomment-173267692
        has_ext = "_d.pyd" in self._data['importlib']['machinery']['EXTENSION_SUFFIXES']
        if with_debug or (with_debug is None and (has_refcount or has_ext)):
            debug = "d"
        if py_version < (3, 8):
            with_pymalloc = self._data['sysconfig']['vars'].get("WITH_PYMALLOC")
            if with_pymalloc or with_pymalloc is None:
                pymalloc = "m"
            if py_version < (3, 3):
                unicode_size = self._data['sysconfig']['vars'].get("Py_UNICODE_SIZE")
                if unicode_size == 4 or (
                        unicode_size is None and self._data['sys']['maxunicode'] == 0x10FFFF
                ):
                    ucs4 = "u"
        elif debug:
            # Debug builds can also load "normal" extension modules.
            # We can also assume no UCS-4 or pymalloc requirement.
            abis.append(f"cp{version}")
        abis.insert(0, f"cp{version}{debug}{pymalloc}{ucs4}", )
        return abis

    @cached_property
    def _compatibility_tag_scorer(self) -> Callable[[str, str, str], Optional[SupportsLessThan]]:

        # some of the code for compatibility tags scoring is based on `packaging::tags` module
        # but instead of output list of tags which requires us to know external information about the version history of
        # gcc, mac, musl, etc. it creates a score function that given a tag will score it so that
        # the user can sort given this score and choose the preferable distribution

        # first we will precompute anything possible:

        my_ipt = self.interpreter_name
        version_rx = re.compile("[0-9_]+")
        my_ipt_version = tuple(self.interpreter_version[:2])
        if len(my_ipt_version) < 2:
            my_ipt_version += (0,)

        if my_ipt == 'cp':
            my_abis = self._compute_cpython_abis(my_ipt_version) if len(my_ipt_version) > 1 else []
            if _abi3_applies(my_ipt_version):
                my_abis.append('abi3')
        else:
            my_abis = list(self._compute_generic_abis())
        my_abis.append('none')
        my_abis.reverse()

        my_platform_type = self._data['platform']['system']
        my_platform = _normalize_string(self._data['sysconfig']['platform'])

        if my_platform_type == "Darwin":
            my_plat_version_str, _, my_cpu_arch = self._data['platform']['mac_ver']
            my_plat_version = tuple(int(it) for it in my_plat_version_str.split("."))[:2]
            my_arch = _mac_arch(my_cpu_arch, self.is_32bit_interpreter)
            my_plat_supported_binary_formats = _mac_binary_formats(my_plat_version, my_arch)
            if my_plat_version[0] >= 11 and my_arch != "x86_64":
                my_plat_supported_binary_formats.append('universal2')
            my_plat_supported_binary_formats.reverse()

            def darwin_score(plt: str) -> Optional[Tuple]:
                plt_components = plt.split('_', 3)
                if plt_components[0] != 'macosx':
                    return None
                _, vmajor_str, vminor_str, binary_format = plt_components
                plat_version = (int(vmajor_str), int(vminor_str))
                if plat_version > my_plat_version:
                    return None
                if (binary_format_index := index_of_or_none(my_plat_supported_binary_formats, binary_format)) is None:
                    return None
                return (binary_format_index + 1, *vdscore(my_plat_version, plat_version))

        elif my_platform_type == "Linux":
            my_linux = my_platform
            if self.is_32bit_interpreter:
                if my_platform == "linux_x86_64":
                    my_linux = "linux_i686"
                elif my_platform == "linux_aarch64":
                    my_linux = "linux_armv7l"

            _, my_arch = my_linux.split("_", 1)

            my_glibc_version = None
            my_musl_version = None
            if (elf := _ElfData.read(Path(self.interpreter_path))) and elf.have_compatible_abi(my_arch):
                my_glibc_version = self._glibc_version.release[:2]
                my_musl_version = elf.compute_musl_version()

            def linux_score(plt: str) -> Optional[Tuple]:
                plt_no_arch = plt[:-(len(my_arch) + 1)]

                if plt == my_linux:
                    return (1, 0, 0)

                if not plt.endswith(f"_{my_arch}"):
                    return None

                if plt_no_arch.startswith("manylinux"):
                    if my_glibc_version is not None:
                        legacy_glibc_version = _LEGACY_MANYLINUX_MAP.get(plt_no_arch)
                        glibc_version = legacy_glibc_version or tuple(int(it) for it in plt_no_arch.split("_")[1:3])
                        if len(glibc_version) != 2 or my_glibc_version < glibc_version:
                            return None
                        return (1, *vdscore(glibc_version, my_glibc_version))
                    return None

                if my_musl_version is not None and plt_no_arch.startswith("musllinux_"):
                    musl_ver = tuple(int(it) for it in plt_no_arch.split("_"))[1:3]
                    if len(musl_ver) != 2 or musl_ver[0] != my_musl_version[0] or musl_ver[1] > my_musl_version[1]:
                        return None
                    return (1, *vdscore(my_musl_version, musl_ver))

                return None

        def vdscore(v1: Tuple, v2: Tuple) -> Tuple:
            if v1[0] == v2[0]:
                return 1, 1 / (1 + abs(v1[1] - v2[1]))
            return 1 / (1 + abs(v1[0] - v2[0])), 0

        # precomputation done. now for creating the scorer

        def scorer(ipt: str, abi: str, plt: str) -> Optional[SupportsLessThan]:
            score_components = []

            # check abi compatibility
            if (abi_index := index_of_or_none(my_abis, abi)) is None:
                return None
            score_components.append(abi_index)

            # check platform
            # platform scores looks like:
            # (binary_format or os score , identifying_component_major_score, identifying_component_minor_score)
            if plt == 'any':
                score_components.extend((0, 0, 0))
            elif plt == my_platform:
                score_components.extend((1, 0, 0))
            elif my_platform_type == "Darwin":
                if (dscore := darwin_score(plt)) is None:
                    return None
                score_components.extend(dscore)
            elif my_platform_type == "Linux":
                if (lscore := linux_score(plt)) is None:
                    return None
                score_components.extend(lscore)
            else:
                return None

            # check compatible interpreter name
            if ipt.startswith('py') and version_rx.fullmatch(ipt[2:]):
                ipt_version_nodot_str = ipt[2:]
                score_components.append(0)
            elif ipt.startswith(my_ipt) and version_rx.fullmatch(ipt[len(my_ipt):]):
                ipt_version_nodot_str = ipt[len(my_ipt):]
                score_components.append(1)
            else:
                return None

            # check compatible interpreter version
            if len(ipt_version_nodot_str) == 1:
                ipt_version = (int(ipt_version_nodot_str), my_ipt_version[1])
            elif '_' in ipt_version_nodot_str:
                ipt_version = tuple(int(p) for p in ipt_version_nodot_str.split("_"))[:2]
            else:
                ipt_version = (int(ipt_version_nodot_str[0]), int(ipt_version_nodot_str[1:]))

            if my_ipt_version[0] != ipt_version[0] or my_ipt_version[1] < ipt_version[1]:
                return None

            score_components.append(vdscore(my_ipt_version, ipt_version)[1])

            return tuple(score_components)

        return scorer

    @classmethod
    def compute(cls, interpreter_path: Path) -> "EnvironmentIntrospection":
        with temp_dir() as tdir:
            intospection_script_path = tdir / 'introspect.py'
            intospection_script_path.write_text(_INTROSPECTION_CODE)

            proc_result = subprocess.run([str(interpreter_path.absolute()), str(intospection_script_path.absolute())],
                                         capture_output=True)
            proc_result.check_returncode()
            output = proc_result.stdout.decode()
            return cls(data=json.loads(output))

    @classmethod
    def load_or_compute(
            cls, path: Path, interpreter_path: Path,
            save_if_recomputed: bool = True) -> "EnvironmentIntrospection":

        data = json.loads(path.read_text()) if path.exists() else {}
        current_introspection_sentinal = hash((_INTROSPECTION_CODE, str(interpreter_path.absolute())))
        if (introspection_sentinal := data.get('introspection')) \
                and introspection_sentinal == current_introspection_sentinal:
            return cls(data=data)

        result = cls.compute(interpreter_path)
        if save_if_recomputed:
            data = result._data
            data['introspection'] = current_introspection_sentinal
            path.parent.mkdir(exist_ok=True, parents=True)
            path.write_text(json.dumps(data))

        return result


def _mac_arch(arch: str, is_32bit: bool) -> str:
    if not is_32bit:
        return arch

    if arch.startswith("ppc"):
        return "ppc"

    return "i386"


INTERPRETER_SHORT_NAMES = {
    "python": "py", "cpython": "cp", "pypy": "pp",
    "ironpython": "ip", "jython": "jy",
}


def _mac_binary_formats(version: Tuple[int, ...], cpu_arch: str) -> List[str]:
    formats = [cpu_arch]
    if cpu_arch == "x86_64":
        if version < (10, 4):
            return []
        formats.extend(["intel", "fat64", "fat32"])

    elif cpu_arch == "i386":
        if version < (10, 4):
            return []
        formats.extend(["intel", "fat32", "fat"])

    elif cpu_arch == "ppc64":
        if version > (10, 5) or version < (10, 4):
            return []
        formats.append("fat64")

    elif cpu_arch == "ppc":
        if version > (10, 6):
            return []
        formats.extend(["fat32", "fat"])

    if cpu_arch in {"arm64", "x86_64"}:
        formats.append("universal2")

    if cpu_arch in {"x86_64", "i386", "ppc64", "ppc", "intel"}:
        formats.append("universal")

    return formats


def _abi3_applies(python_version: Tuple) -> bool:
    return len(python_version) > 1 and python_version >= (3, 2)


def _normalize_string(string: str) -> str:
    return string.replace(".", "_").replace("-", "_")


_EM_ARM = 40
_EM_386 = 3
_EF_ARM_ABIMASK = 0xFF000000
_EF_ARM_ABI_VER5 = 0x05000000
_EF_ARM_ABI_FLOAT_HARD = 0x00000400
_PT_INTERP = 3
_elf_header_prefix_struct = Struct("4s5B7x")


class _ElfStructs:
    # noinspection PyPep8Naming
    def __init__(self, is_32_bit: bool, is_little_endian: bool):
        L = "L" if is_32_bit else "Q"
        e = "<" if is_little_endian else ">"

        self.header_suffix = Struct(f"{e}2HI3{L}I6H")
        self.program_header = Struct(f"{e}2I4{L}")


def _unpack(fd: IO, struct: Struct) -> Tuple:
    return struct.unpack(fd.read(struct.size))


@dataclass
class _ElfData:
    is_32_bit: bool
    is_little_endian: bool
    machine: int
    flags: int
    interpreter: Optional[str]

    def __str__(self):
        return f"ELF(is_32_bit={self.is_32_bit}, is_little_endian={self.is_little_endian}, " \
               f"machine={self.machine:x}, flags={self.flags:b}, " \
               f"interpreter={self.interpreter})"

    def _is_linux_armhf(self) -> bool:
        result = self.is_32_bit & self.is_little_endian and self.machine == _EM_ARM
        result &= (self.flags & _EF_ARM_ABIMASK) == _EF_ARM_ABI_VER5
        result &= (self.flags & _EF_ARM_ABI_FLOAT_HARD) == _EF_ARM_ABI_FLOAT_HARD
        return result

    def _is_linux_i686(self) -> bool:
        return self.is_32_bit and self.is_little_endian and self.machine == _EM_386

    def compute_musl_version(self) -> Optional[Tuple[int, ...]]:
        if not self.interpreter:
            return None

        proc = subprocess.run([self.interpreter], stderr=subprocess.PIPE, universal_newlines=True)
        lines = [n for n in (n.strip() for n in proc.stderr.splitlines()) if n]
        if len(lines) >= 2 and lines[0].startswith('musl'):
            try:
                return tuple(int(it) for it in lines[1].split()[1].split("."))[:2]
            except:  # noqa
                return None

    def have_compatible_abi(self, arch: str) -> bool:
        if arch == "armv7l":
            return self._is_linux_armhf()
        if arch == "i686":
            return self._is_linux_i686()
        return arch in {"x86_64", "aarch64", "ppc64", "ppc64le", "s390x"}

    @classmethod
    def read(cls, file: Path) -> Optional["_ElfData"]:
        with file.open('rb') as fd:
            uprefix = _unpack(fd, _elf_header_prefix_struct)

            if uprefix[0] != b'\x7fELF':
                return None

            is_32_bit, is_little_endian = (uprefix[1] == 1), (uprefix[2] == 1)
            structs = _ElfStructs(is_32_bit, is_little_endian)

            usuffix = _unpack(fd, structs.header_suffix)
            machine, flags = usuffix[1], usuffix[6]
            ph_offset, n_ph, ph_size = usuffix[4], usuffix[9], usuffix[8]

            interpreter = None
            for i in range(n_ph):
                fd.seek(ph_offset + i * ph_size)
                header = _unpack(fd, structs.program_header)
                if header[0] == _PT_INTERP:
                    sz = header[5]
                    fd.seek(header[2])
                    interpreter = os.fsdecode(fd.read(sz)).strip("\0")
                    break

            return _ElfData(is_32_bit, is_little_endian, machine, flags, interpreter)


_LEGACY_MANYLINUX_MAP: Dict[str, Tuple[int, ...]] = {
    "manylinux2014": (2, 17),
    "manylinux2010": (2, 12),
    "manylinux1": (2, 5),
}



import json
import threading
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent
from typing import Optional, Dict, Set, Literal, Any, List, TYPE_CHECKING

from pkm.api.dependencies.dependency import Dependency
from pkm.api.distributions.build_monitors import BuildPackageMonitoredOp, BuildPackageHookExecutionEvent
from pkm.api.packages.package import PackageDescriptor
from pkm.api.projects.pyproject_configuration import BuildSystemConfig
from pkm.utils.files import temp_dir

if TYPE_CHECKING:
    from pkm.api.environments.environment import Environment
    from pkm.api.projects.project import Project


class BuildError(IOError):

    def __init__(self, msg: str, missing_hook: bool = False) -> None:
        super().__init__(msg)
        self.missing_hook = missing_hook


_ongoing_builds: Dict[int, Set[PackageDescriptor]] = defaultdict(set)


@contextmanager
def _cycle_detection(project: "Project"):
    ongoing_builds = _ongoing_builds[threading.current_thread().ident]
    if project.descriptor in ongoing_builds:
        raise BuildError(f"cycle detected involving: {ongoing_builds}")

    ongoing_builds.add(project.descriptor)
    try:
        yield
    finally:
        ongoing_builds.remove(project.descriptor)
        if not ongoing_builds:
            del _ongoing_builds[threading.current_thread().ident]


def build_sdist(project: "Project", target_dir: Optional[Path] = None,
                target_env: Optional["Environment"] = None) -> Path:
    """
    build a source distribution from this project
    :param project: the project to build
    :param target_dir: the directory to put the created archive in
    :param target_env: the environment to build this distribution for
    :return: the path to the created archive
    """

    from pkm.api.environments.environment_builder import EnvironmentBuilder
    target_dir = target_dir or (project.directories.dist / str(project.version))
    target_env = target_env or project.attached_environment

    target_dir.mkdir(exist_ok=True, parents=True)

    dist = 'sdist'
    with BuildPackageMonitoredOp(project.descriptor, dist) as mop, temp_dir() as tdir, _cycle_detection(project):

        pyproject = project.config
        buildsys: BuildSystemConfig = pyproject.build_system
        build_packages_repo = project.attached_repository

        build_env = EnvironmentBuilder.create(tdir / 'venv', target_env.interpreter_path)

        if buildsys.requirements:
            build_env.install(buildsys.requirements, build_packages_repo)
        if buildsys.backend_path:
            build_env.install_link('build_backend', [project.path / pth for pth in buildsys.backend_path])

        # start build life-cycle:
        # 1. check for sdist extra requirements
        command = 'get_requires_for_build_sdist'
        mop.notify(BuildPackageHookExecutionEvent(project.descriptor, command))

        extra_requirements = _exec_build_cycle_script(project, build_env, buildsys, command, [None])

        if extra_requirements.status == 'success':
            if requirements := [Dependency.parse(d) for d in extra_requirements.result]:
                build_env.install(requirements, build_packages_repo)

        # 2. build the sdist
        command = 'build_sdist'
        mop.notify(BuildPackageHookExecutionEvent(project.descriptor, command))
        sdist_output = _exec_build_cycle_script(
            project, build_env, buildsys, command, [str(target_dir), None])

        if sdist_output.status == 'success':
            return target_dir / sdist_output.result
        raise BuildError("build backend did not produced expected sdist", True)


def build_wheel(project: "Project", target_dir: Optional[Path] = None, only_meta: bool = False,
                editable: bool = False, target_env: Optional["Environment"] = None) -> Path:
    """
    build a wheel distribution from this project
    :param project: the project to build
    :param target_dir: directory to put the resulted wheel in
    :param only_meta: if True, only builds the dist-info directory otherwise the whole wheel
    :param editable: if True, a wheel for editable install will be created
    :param target_env: the environment that this build should be compatible with
    :return: path to the built artifact (directory if only_meta, wheel archive otherwise)
    """

    from pkm.api.environments.environment_builder import EnvironmentBuilder
    target_dir = target_dir or (project.directories.dist / str(project.version))
    interpreter_path = target_env.interpreter_path if target_env else project.attached_environment.interpreter_path

    target_dir.mkdir(exist_ok=True, parents=True)

    dist = 'editable_wheel' if editable else 'metadata' if only_meta else 'wheel'
    with BuildPackageMonitoredOp(project.descriptor, dist) as mop, temp_dir() as tdir, _cycle_detection(project):

        pyproject = project.config
        buildsys: BuildSystemConfig = pyproject.build_system
        build_packages_repo = project.attached_repository

        build_env = EnvironmentBuilder.create(tdir / 'venv', interpreter_path)

        if buildsys.requirements:
            build_env.install(buildsys.requirements, build_packages_repo)
        if buildsys.backend_path:
            build_env.install_link('build_backend', [project.path / pth for pth in buildsys.backend_path])

        # start build life-cycle:
        # 1. check for wheel extra requirements
        command = 'get_requires_for_build_editable' \
            if editable else 'get_requires_for_build_wheel'
        mop.notify(BuildPackageHookExecutionEvent(project.descriptor, command))

        extra_requirements = _exec_build_cycle_script(project, build_env, buildsys, command, [None])

        if extra_requirements.status == 'success':
            build_env.install(
                [Dependency.parse(d) for d in extra_requirements.result],
                build_packages_repo)

        if only_meta:
            # 2. try to build metadata only
            command = 'prepare_metadata_for_build_wheel'
            mop.notify(BuildPackageHookExecutionEvent(project.descriptor, command))
            dist_info_output = _exec_build_cycle_script(
                project, build_env, buildsys, command,
                [str(target_dir), None])
            if dist_info_output.status == 'success':
                return target_dir / dist_info_output.result
            raise BuildError("build backend did not produced wheel metadata", True)

        # 3. build the wheel
        command = 'build_editable' if editable else 'build_wheel'
        mop.notify(BuildPackageHookExecutionEvent(project.descriptor, command))
        wheel_output = _exec_build_cycle_script(
            project, build_env, buildsys, command, [str(target_dir), None, None])

        if wheel_output.status == 'success':
            return target_dir / wheel_output.result
        raise BuildError("build backend did not produced expected wheel")


@dataclass
class _BuildCycleResult:
    status: Literal['success', 'undefined_hook']
    result: Any


def _exec_build_cycle_script(
        project: "Project", env: "Environment", buildsys: BuildSystemConfig, hook: str,
        arguments: List[Any]) -> _BuildCycleResult:
    with temp_dir() as tdir_path:
        build_backend_parts = buildsys.build_backend.split(":")
        build_backend_import = build_backend_parts[0]
        build_backend = 'build_backend' + (f".{build_backend_parts[1]}" if len(build_backend_parts) > 1 else "")
        output_path = tdir_path / 'output'
        output_path_str = str(output_path.absolute()).replace('\\', '\\\\')

        script = f"""
            import json

            def ret(status, result):
                out = open('{output_path_str}', 'w+')
                out.write(json.dumps({{'status': status, 'result': result}}))
                out.close()
                exit(0)

            try:
                import {build_backend_import} as build_backend
                if not hasattr({build_backend}, '{hook}'):
                    ret('undefined_hook', None)
                else:
                    result = {build_backend}.{hook}({', '.join(repr(arg) for arg in arguments)})
                    ret('success', result)
            except Exception:
                import traceback
                traceback.print_exc()
                ret('fail', traceback.format_exc())
        """

        script_path = tdir_path / 'execution.py'
        script_path.write_text(dedent(script))
        process_results = env.run_proc([str(env.interpreter_path), str(script_path)], cwd=project.path)
        if process_results.returncode != 0:
            raise BuildError(
                f"PEP517 build cycle execution failed.\n"
                f"Project: {project.name} {project.version}\n"
                f"Build backend: {build_backend}\n"
                f"Hook: {hook}\n"
                f"Resulted in exit code: {process_results.returncode}", False)

        result = _BuildCycleResult(**json.loads((tdir_path / 'output').read_text()))
        if result.status == 'fail':
            raise BuildError(
                f"PEP517 build cycle execution failed.\n"
                f"Project: {project.name} {project.version}\n"
                f"Build backend: {build_backend}\n"
                f"Hook: {hook}\n"
                f"Resulted in exception:\n{result.result}", False)
        return result



from typing import TypeVar, MutableMapping, Optional, Callable, Dict, Mapping

from pkm.utils.types import SupportHashCode

_K = TypeVar("_K", bound=SupportHashCode)
_V = TypeVar("_V")
_M = TypeVar("_M", bound=MutableMapping)
_IM = TypeVar("_IM", bound=Mapping)
_MISSING = object()


def put_if_absent(d: MutableMapping[_K, _V], key: _K, value: _V) -> bool:
    """
    sets `d[key]=value` if `key not in d`
    :param d: the dict to mutate
    :param key: the key to assign
    :param value: the value to set if `key not in d`
    :return: true if `value` was set (`key` was not in `d`), false otherwise.
    """
    if key in d:
        return False
    d[key] = value
    return True


def get_or_put(d: MutableMapping[_K, _V], key: _K, value_provider: Callable[[], _V]) -> _V:
    """
    :param d: the dict to look in
    :param key: the key to get
    :param value_provider: function providing a new value for `d[key]` in the case where no such value already exists
    :return: `d[key]` if such value exists otherwise set `d[key] = value_provider()` and then returns `d[key]`
    """
    if (value := d.get(key, _MISSING)) is _MISSING:
        d[key] = value = value_provider()

    return value


def get_or_compute(d: Mapping[_K, _V], key: _K, value_provider: Callable[[], _V]) -> _V:
    """
    :param d: the dict to look in
    :param key: the key to get
    :param value_provider: function providing a new value for `d[key]` in the case where no such value already exists
    :return: `d[key]` if such value exists otherwise value_provider(), does not modifies `d`.
    """
    if (value := d.get(key, _MISSING)) is _MISSING:
        return value_provider()

    return value


def get_or_raise(d: MutableMapping[_K, _V], key: _K, err_provider: Callable[[], Exception]) -> _V:
    """
    attempt to get a value from `d` with the given `key`, if no such value exists raise the
    exception returned by `err_provider`
    :param d: the mapping to look in
    :param key: the key to look for
    :param err_provider: callable that may be called to provide an exception instance in case where `key` is not in `d`
    :return: the value `d[key]`
    """

    if (result := d.get(key, _MISSING)) is _MISSING:
        raise err_provider()

    return result


def without_keys(d: Mapping[_K, _V], *keys: _K) -> Dict[_K, _V]:
    """
    creates a new dict with the same content as `d` minus the keys that are in `keys`
    :param d: the mapping to clone
    :param keys: the keys to leave out
    :return: the newly created dict
    """
    remove_set = set(keys)
    return {k: v for k, v in d.items() if k not in remove_set}


def remove_none_values(d: "_M[_K, _V]") -> "_M[_K, _V]":
    """
    remove from `d` all items with value equals to None and return back `d`
    :param d: the dict to remove from
    :return: `d` after the required changes
    """
    return remove_by_value(d)


def remove_by_value(
        d: "_M[_K, _V]", value: Optional[_V] = None,
        match: Optional[Callable[[_V], bool]] = None) -> "_M[_K, _V]":
    """
    remove items from [d] if they either match the given [match] function
    or if no [match] function given, if they are equals to the given [value]

    :param d: the dict to remove items from
    :param value: if supplied, the value to remove
    :param match: if supplied, matcher for the values to remove
    :return: [d]
    """

    if match:
        keys_to_remove = [k for k, v in d.items() if match(v)]
    else:
        keys_to_remove = [k for k, v in d.items() if v == value]

    for k in keys_to_remove:
        del d[k]

    return d


def udict_hash(d: "_IM[_K, _V]", key_hash: Callable[[_K], int] = hash, value_hash: Callable[[_V], int] = hash):
    """
    unordered dict hash - computes a hash value for the given dict, the hash will be the same for all dicts with
    the same elements without considering their insertion order
    :param d: the dict to compute hash to
    :param key_hash: a function that can compute hash for the dict keys
    :param value_hash: a function that can compute hash for the dict values
    :return: the computed hashcode
    """
    result = 0
    for k, v in d.items():
        kvh = (7 * 31 + key_hash(k)) * 31 + value_hash(v)
        result = result ^ kvh
    return result



from abc import abstractmethod, ABC
from base64 import b64encode
from dataclasses import dataclass
from pathlib import Path
from typing import List, Union, Optional, Tuple, Any, Protocol

from pkm.api.dependencies.dependency import Dependency
from pkm.api.packages.package import Package
from pkm.api.packages.package_metadata import PackageMetadata
from pkm.api.versions.version_specifiers import AnyVersion
from pkm.utils.iterators import partition


class Repository(Protocol):

    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @abstractmethod
    def match(self, dependency: Union[Dependency, str], check_prereleases: bool = True) -> List[Package]:
        """
        :param dependency: the dependency to match (or a pep508 string representing it)
        :param check_prereleases: whether or not to check pre-releases according to pep440 rules.
              if True, will only output pre-releases if the dependency version specifier is a
              pre-release or all the versions matching the dependency are pre-releases.
              Otherwise, will output all pre-releases matching the dependency

        :return: list of all the packages in this repository that match the given `dependency`
        """

    def list(self, package_name: str) -> List[Package]:
        """
        :param package_name: the package to match 
        :return: list of all the packages that match the given `package_name`
        """
        return self.match(Dependency(package_name, AnyVersion))

    @property
    @abstractmethod
    def publisher(self) -> Optional["RepositoryPublisher"]:
        """
        :return: if this repository is 'publishable' returns its publisher
        """


# noinspection PyMethodMayBeStatic
class AbstractRepository(Repository, ABC):

    def __init__(self, name: str):
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    def match(self, dependency: Union[Dependency, str], check_prereleases: bool = True) -> List[Package]:
        if isinstance(dependency, str):
            dependency = Dependency.parse(dependency)

        matched = self._do_match(dependency)
        filtered = self._filter_prereleases(matched, dependency) if check_prereleases else matched
        return self._sort_by_priority(dependency, filtered)

    @property
    def publisher(self) -> Optional["RepositoryPublisher"]:
        return None

    def _filter_prereleases(self, packages: List[Package], dependency: Dependency) -> List[Package]:
        if dependency.version_spec.allows_pre_or_dev_releases():
            return packages
        pre_release, rest = partition(packages, lambda it: it.version.is_pre_or_dev_release())
        return rest or packages

    def _sort_by_priority(self, dependency: Dependency, packages: List[Package]) -> List[Package]:
        """
        sorts `matches` by the required priority
        :param dependency: the dependency that resulted in the given `packages`
        :param packages: the packages that were the result `_do_match(dependency)`
        :return: sorted packages by priority (first is more important than last)
        """
        packages.sort(key=lambda it: it.version, reverse=True)
        return packages

    @abstractmethod
    def _do_match(self, dependency: Dependency) -> List[Package]:
        """
        IMPLEMENTATION NOTICE:
            1. do not try to filter pre-releases,
               it is handled for you in the `match` method that call this one.
            2. the matching should only be done against the package name and the version spec
               since the environment markers should not be considered in the level of the repository

        :param dependency: the dependency to match
        :return: list of all the packages in this repository that match the given `dependency` version spec
        """


class RepositoryPublisher:
    def __init__(self, repository_name: str):
        self.repository_name = repository_name

    # noinspection PyMethodMayBeStatic
    def required_authentication_fields(self) -> List[str]:
        return ['username', 'password']

    @abstractmethod
    def publish(self, auth: "Authentication", package_meta: PackageMetadata, distribution: Path):
        """
        publish a `distribution` belonging to the given `package_meta` into the repository (registering it if needed)
        :param auth: authentication object filled with the fields that were
                     returned by the method `required_authentication_fields`
        :param package_meta: metadata for the package that this distribution belongs to
        :param distribution: the distribution archive (e.g., wheel, sdist)
        """


@dataclass(frozen=True, eq=True)
class Authentication:
    username: str
    password: str

    def as_basic_auth_header(self) -> Tuple[str, str]:
        return 'Authorization', f'Basic {b64encode(f"{self.username}:{self.password}".encode()).decode("ascii")}'


class RepositoryBuilder(ABC):

    def __init__(self, name: str):
        self.name = name

    @abstractmethod
    def build(self, name: Optional[str], packages: Optional[List[str]],
              **kwargs: Any) -> Repository:
        """
        build a new repository instance using the given `kwargs`
        :param name: name for the created repository
        :param packages: list of packages the user ask the repository to be limited to
                (or None if no such request was made)
        :param kwargs: arguments for the instance creation, may be defined by derived classes
        :return: the created instance
        """



from typing import TypeVar, Callable, Iterable, Sequence, Optional, Any, List

from pkm.utils.types import SupportHashCode, SupportsLessThan

_T = TypeVar("_T")
_U = TypeVar("_U")
_K = TypeVar("_K", bound=SupportHashCode)


def subiter(sq: Sequence[_T], offset: int = 0, length: Optional[int] = None) -> Iterable[_T]:
    if length is None:
        length = len(sq) - offset

    for i in range(offset, offset + length):
        yield sq[i]


def startswith(sq: Sequence[_T], prefix: Sequence[_T]) -> bool:
    if len(prefix) > len(sq):
        return False

    return all(s == p for s, p in zip(sq, prefix))


def argmin(seq: Sequence[_T], key: Optional[Callable[[_T], SupportsLessThan]]) -> int:
    if key is None:
        okey = seq.__getitem__
    else:
        def okey(i: int):
            return key(seq[i])

    return min(range(len(seq)), key=okey)


def argmax(seq: Sequence[_T], key: Optional[Callable[[_T], SupportsLessThan]]) -> int:
    if key is None:
        okey = seq.__getitem__
    else:
        def okey(i):
            return key(seq[i])

    return max(range(len(seq)), key=okey)


def index_of_or_none(seq: Sequence[_T], value: Any) -> Optional[int]:
    """
    :param seq: the sequence to search in
    :param value: the value to search for
    :return: the index of `value` in `seq` or `None` if `value not in seq`
    """
    try:
        return seq.index(value)
    except ValueError:
        return None


def single_or_raise(seq: Sequence[_T]) -> _T:
    """
    :param seq: the seq to access
    :return: the first element of this sequence if `len(seq) == 1`  otherwise raise `ValueError`
    """
    if (l := len(seq)) != 1:
        raise ValueError(f"expecting single element, found: {l}")
    return seq[0]


def get_or_default(seq: Sequence[_T], index: int, default: _T) -> _T:
    """
    :param seq: the sequence to access
    :param index: the index to get
    :param default: the value to return if `-len(seq) > index > len(seq)`
    :return: the value of `seq[index]` if such exists (no IndexError had been thrown upon accessing index),
             otherwise returns the `default` value
    """
    if -len(seq) > index > len(seq):
        return default
    return seq[index]


def oseq_hash(seq: List[_T], item_hash: Callable[[_K], int] = hash):
    """
    ordered sequence hash - computes a hash value for the given seq
    :param seq: the sequence to compute hash to
    :param item_hash: a function that can compute hash for the sequence items
    :return: the computed hashcode
    """
    result = 7
    for item in seq:
        result = result * 31 + item_hash(item)
    return result


def strs(seq: Sequence[_T]) -> List[str]:
    """
    :param seq: the sequence to extract items from
    :return: list containing `str(item)` for each item in `seq`
    """
    return [str(it) for it in seq]



from dataclasses import dataclass

from pkm.api.packages.package import PackageDescriptor
from pkm.utils.monitors import MonitoredOperation, MonitoredEvent


@dataclass
class BuildPackageMonitoredOp(MonitoredOperation):
    package: PackageDescriptor
    distribution: str  # can be: wheel, metadata, editable_wheel, sdist


@dataclass
class BuildPackageHookExecutionEvent(MonitoredEvent):
    package: PackageDescriptor
    hook: str



from __future__ import annotations

import itertools
from typing import TypeVar, Generic, Iterator, Callable, Optional, List, Iterable

_T = TypeVar("_T")
_U = TypeVar("_U")


def seq(iterator: Iterable[_T]) -> Seq[_T]:
    return Seq(iter(iterator))


class Seq(Generic[_T], Iterator[_T]):

    def __init__(self, iterator: Iterator[_T]):
        self._iter = iterator

    def __next__(self) -> _T:
        return next(self._iter)

    def map(self, mapper: Callable[[_T], _U]) -> Seq[_U]:
        return Seq(mapper(it) for it in self._iter)

    def flatmap(self, mapper: Callable[[_T], Iterable[_U]]) -> Seq[_U]:
        return Seq((item for collection in self.map(mapper) for item in collection))

    def map_not_none(self, mapper: Callable[[_T], _U]) -> Seq[_U]:
        return Seq(m for it in self._iter if (m := mapper(it)) is not None)

    def filter(self, accept: Callable[[_T], bool]) -> Seq[_T]:
        return Seq(it for it in self._iter if accept(it))

    def chain(self, other: Iterator[_T]):
        return Seq(itertools.chain(self._iter, other))

    def first_or(self, default: _T) -> _T:
        try:
            return next(self._iter)
        except StopIteration:
            return default

    def first_or_none(self) -> Optional[_T]:
        return self.first_or(None)

    def find_or(self, accept: Callable[[_T], bool], default: _T) -> _T:
        return self.filter(accept).first_or(default)

    def find_or_none(self, accept: Callable[[_T], bool]) -> Optional[_T]:
        return self.find_or(accept, None)

    def limit(self, amount: int) -> Seq[_T]:
        return Seq(itertools.islice(self._iter, amount))

    def for_each(self, op: Callable[[_T], None]) -> None:
        for it in self._iter:
            op(it)

    def to_list(self, into: Optional[List[_T]] = None) -> List[_T]:
        if into:
            into.extend(self._iter)
        else:
            into = list(self._iter)

        return into



import hashlib
import io
import re
from base64 import urlsafe_b64encode
from dataclasses import dataclass
from pathlib import Path
from typing import Protocol, IO, Union

from pkm.utils.io_streams import chunks

_SIG_DELIM_RX = re.compile("\\s*=\\s*")


class HashDigester(Protocol):
    def digest(self) -> bytes:
        ...

    def hexdigest(self) -> str:
        ...

    def update(self, __data: Union[bytes, bytearray, memoryview]) -> None:
        ...


def stream(hd: HashDigester, file: Union[IO, Path], chunk_size: int = io.DEFAULT_BUFFER_SIZE):
    """
    streams a file into `hd` (via `hd.update(chunk)`) using chunks of the given `chunk_size`
    :param hd: the hash digester to stream into
    :param file: the file to stream
    :param chunk_size: the size of the chunk to use, defaults to `io.DEFAULT_BUFFER_SIZE`
    """
    if isinstance(file, Path):
        with file.open('rb') as source_fd:
            stream(hd, source_fd, chunk_size)
    else:
        for chunk in chunks(file, chunk_size):
            hd.update(chunk)


@dataclass
class HashSignature:
    hash_type: str
    hash_value: str

    def _encode_hash(self, hashd: HashDigester) -> str:
        return type(self).encode_hash(hashd)

    def validate_against(self, file: Path) -> bool:
        if not hasattr(hashlib, self.hash_type):
            raise KeyError(f"Cannot validate archive, Unsupported Hash {self.hash_type}")

        hash_computer = getattr(hashlib, self.hash_type)()
        stream(hash_computer, file)
        computed_hash = self._encode_hash(hash_computer)
        return computed_hash == self.hash_value

    def __str__(self):
        return f"{self.hash_type}={self.hash_value}"

    def __repr__(self):
        return f"HashSignature({self})"

    @classmethod
    def encode_hash(cls, hashd: HashDigester):
        return hashd.hexdigest()

    @classmethod
    def parse_hex_encoded(cls, signature: str) -> "HashSignature":
        parts = _SIG_DELIM_RX.split(signature)
        if len(parts) != 2:
            raise ValueError('unsupported signature, expecting format <hash_type>=<hash_value>')

        return HashSignature(*parts)

    @classmethod
    def parse_urlsafe_base64_nopad_encoded(cls, signature: str) -> "HashSignature":
        parts = _SIG_DELIM_RX.split(signature)
        if len(parts) != 2:
            raise ValueError(f'unsupported signature, expecting format <hash_type>=<hash_value>, got: {signature}')

        return _UrlsafeBase64NopadHashSignature(*parts)

    @classmethod
    def create_urlsafe_base64_nopad_encoded(cls, hash_function: str, file: Path) -> "HashSignature":
        hashd = getattr(hashlib, hash_function)()
        stream(hashd, file)
        encoded = _UrlsafeBase64NopadHashSignature.encode_hash(hashd)
        return _UrlsafeBase64NopadHashSignature(hash_function, encoded)


class _UrlsafeBase64NopadHashSignature(HashSignature):

    @classmethod
    def encode_hash(cls, hashd: HashDigester):
        return urlsafe_b64encode(hashd.digest()).decode("latin1").rstrip("=")



from __future__ import annotations

from dataclasses import dataclass, replace
from typing import Optional, List, TYPE_CHECKING

from pkm.api.dependencies.env_markers import PEP508EnvMarkerParser, EnvironmentMarker
from pkm.api.versions.version_parser import VersionParser
from pkm.api.versions.version_specifiers import VersionSpecifier, AnyVersion
from pkm.utils.parsers import SimpleParser

if TYPE_CHECKING:
    from pkm.api.environments.environment import Environment
    from pkm.api.packages.package import PackageDescriptor


@dataclass(frozen=True, eq=True)
class Dependency:
    """
    represents a package dependency (`package a` depends on `package b`)
    where the fields: `package_name` refer to the dependant package, `version_spec` refer to the version requirement,
    `extras` refers to the required package extras and finally, env marker is a constraint for the environments that
    this dependency is applicable for
    """

    package_name: str
    version_spec: VersionSpecifier = AnyVersion
    extras: Optional[List[str]] = None
    env_marker: Optional[EnvironmentMarker] = None

    def is_applicable_for(self, env: "Environment", extras: List[str]) -> bool:
        return self.env_marker is None or self.env_marker.evaluate_on(env, extras or [])

    def with_extras(self, extras: Optional[List[str]]) -> Dependency:
        return replace(self, extras=extras)

    def __str__(self):
        extras_str = f"[{','.join(self.extras)}]" if self.extras else ''

        version_str = str(self.version_spec)

        marker_str = f";{self.env_marker}" if self.env_marker else ''

        return f"{self.package_name}{extras_str} {version_str}{marker_str}"

    def __repr__(self):
        return f"Dependency({self})"

    @classmethod
    def parse(cls, text: str) -> "Dependency":
        return PEP508DependencyParser(text).read_dependency()


class PEP508DependencyParser(SimpleParser):

    def _read_extras(self) -> List[str]:
        self.match_or_err('[', 'expecting extras start ([)')
        extras: List[str] = []

        while self.is_not_empty():
            self.match_ws()
            extras.append(self._read_identifier())
            self.match_ws()
            if not self.match(','):
                break

        self.match_or_err(']', 'expecting extras end (])')

        return extras

    def _read_version_spec(self) -> VersionSpecifier:
        return self.subparser(VersionParser).read_specifier()

    def _read_identifier(self) -> str:
        if self.peek().isalpha():
            return self.until(lambda i, t: not t[i].isalnum() and t[i] not in '_-.')
        self.raise_err('expecting identifier')

    def read_emarker(self) -> EnvironmentMarker:
        return self.subparser(PEP508EnvMarkerParser).read_marker()

    def read_dependency(self) -> Dependency:
        from pkm.api.packages.package import PackageDescriptor

        self.match_ws()
        package_name = PackageDescriptor.normalize_name(self._read_identifier())
        self.match_ws()

        extras: Optional[List[str]] = None
        if self.peek() == '[':
            extras = self._read_extras()
            self.match_ws()

        version_spec: VersionSpecifier = AnyVersion
        if self.peek() != ';':
            version_spec = self._read_version_spec()
            self.match_ws()

        env_marker: Optional[EnvironmentMarker] = None
        if self.match(";", '') and self.is_not_empty():
            env_marker = self.read_emarker()

        return Dependency(package_name, version_spec, extras=extras, env_marker=env_marker)



import shutil
from pathlib import Path
from typing import Optional, List, Dict, Iterable, TYPE_CHECKING, Iterator

from pkm.api.dependencies.dependency import Dependency
from pkm.api.distributions.distinfo import DistInfo, RequestedPackageInfo
from pkm.api.packages.package import Package, PackageDescriptor
from pkm.api.packages.package_metadata import PackageMetadata
from pkm.api.versions.version_specifiers import SpecificVersion
from pkm.utils.files import is_empty_directory, is_relative_to
from pkm.utils.properties import cached_property, clear_cached_properties

if TYPE_CHECKING:
    from pkm.api.environments.environment import Environment
    from pkm.api.packages.package_installation import PackageInstallationTarget


class SitePackages:
    def __init__(self, env: "Environment", purelib: Path, platlib: Path):

        self._purelib = purelib
        self._platlib = platlib
        self.env = env

    def all_sites(self) -> Iterator[Path]:
        yield self._purelib
        if self._platlib != self._purelib:
            yield self._platlib

    def __eq__(self, other):
        return isinstance(other, SitePackages) \
               and other._purelib == self._purelib \
               and other._platlib == self._platlib \
               and other.env.path == self.env.path

    @property
    def purelib_path(self) -> Path:
        return self._purelib

    @property
    def platlib_path(self) -> Path:
        return self._platlib

    @cached_property
    def _name_to_packages(self) -> Dict[str, "InstalledPackage"]:
        result: Dict[str, InstalledPackage] = {}
        self._scan_packages(self._purelib, result)
        self._scan_packages(self._platlib, result)
        return result

    def _scan_packages(self, site: Path, result: Dict[str, "InstalledPackage"]):
        if not site.exists():
            return

        for di in DistInfo.scan(site):
            result[di.package_name] = InstalledPackage(di, self)

    def installed_packages(self) -> Iterable["InstalledPackage"]:
        return self._name_to_packages.values()

    def installed_package(self, package_name: str) -> Optional["InstalledPackage"]:
        return self._name_to_packages.get(PackageDescriptor.normalize_src_package_name(package_name))

    def reload(self):
        clear_cached_properties(self)


def _read_user_request(dist_info: DistInfo, metadata: PackageMetadata) -> Optional[RequestedPackageInfo]:
    if stored_request := dist_info.load_user_requested_info():
        return stored_request
    elif dist_info.is_user_requested():
        return RequestedPackageInfo(Dependency(metadata.package_name, SpecificVersion(metadata.package_version)), False)


class InstalledPackage(Package):

    def __init__(self, dist_info: DistInfo, site: Optional[SitePackages] = None):

        self._dist_info = dist_info
        self.site = site

    @cached_property
    def published_metadata(self) -> Optional["PackageMetadata"]:
        return self._dist_info.load_metadata_cfg()

    @property
    def dist_info(self) -> DistInfo:
        """
        :return: the installed package dist-info
        """
        return self._dist_info

    @cached_property
    def descriptor(self) -> PackageDescriptor:
        meta = self.published_metadata
        return PackageDescriptor(meta.package_name, meta.package_version)

    @cached_property
    def user_request(self) -> Optional[RequestedPackageInfo]:
        """
        :return: information about the dependency that was requested by the user
                 if this package was directly requested by the user or its project
                 otherwise None
        """
        return _read_user_request(self._dist_info, self.published_metadata)

    def _all_dependencies(self, environment: "Environment") -> List["Dependency"]:
        return self.published_metadata.dependencies

    def is_compatible_with(self, env: "Environment") -> bool:
        return self.published_metadata.required_python_spec.allows_version(env.interpreter_version)

    def install_to(
            self, target: "PackageInstallationTarget", user_request: Optional["Dependency"] = None,
            editable: bool = False):
        if target.site_packages != self.site:
            raise NotImplemented()  # maybe re-mark user request?

    def is_in_purelib(self) -> bool:
        """
        :return: True if this package is installed to purelib, False if it is installed into platlib
        """

        return is_relative_to(self.dist_info.path, self.site.purelib_path)

    def uninstall(self):
        """
        uninstall this package from its site packages
        """
        parents_to_check = set()
        for installed_file in self._dist_info.installed_files():
            installed_file.unlink(missing_ok=True)
            parents_to_check.add(installed_file.parent)

        installation_site = self.dist_info.path.parent
        while parents_to_check:
            parent = parents_to_check.pop()  # noqa : pycharm does not know's about set's pop method?

            if parent == installation_site or not is_relative_to(parent, installation_site):
                continue

            if (precompiled := parent / "__pycache__").exists():
                shutil.rmtree(precompiled, ignore_errors=True)

            if is_empty_directory(parent):
                parent.rmdir()
                parents_to_check.add(parent.parent)

        if self._dist_info.path.exists():
            shutil.rmtree(self._dist_info.path)

        if self.site:
            self.site.reload()



from __future__ import annotations

import shutil
from dataclasses import dataclass, replace
from pathlib import Path
from typing import List, Optional, Union, TYPE_CHECKING, Dict, Callable

from pkm.api.distributions.distinfo import DistInfo, RequestedPackageInfo
from pkm.api.environments.environment_builder import EnvironmentBuilder
from pkm.api.packages.package import Package, PackageDescriptor
from pkm.api.packages.package_installation import PackageInstallationTarget
from pkm.api.packages.package_metadata import PackageMetadata
from pkm.api.packages.package_monitors import PackageInstallMonitoredOp
from pkm.api.pkm import pkm
from pkm.api.projects.environments_config import EnvironmentsConfiguration, ENVIRONMENT_CONFIGURATION_PATH, \
    AttachedEnvironmentConfig
from pkm.api.projects.pyproject_configuration import PyProjectConfiguration, PkmDistributionConfig, \
    PKM_DIST_CFG_TYPE_LIB, PkmApplicationConfig, PKM_DIST_CFG_TYPE_CAPP
from pkm.api.repositories.repository import Repository, RepositoryPublisher, Authentication
from pkm.api.versions.version import StandardVersion, Version, NamedVersion
from pkm.api.versions.version_specifiers import VersionRange, SpecificVersion
from pkm.resolution.packages_lock import PackagesLock
from pkm.utils.commons import UnsupportedOperationException
from pkm.utils.files import temp_dir
from pkm.utils.properties import cached_property, clear_cached_properties

if TYPE_CHECKING:
    from pkm.api.projects.project_group import ProjectGroup
    from pkm.api.environments.environment import Environment
    from pkm.api.dependencies.dependency import Dependency


class Project(Package):

    def __init__(self, pyproject: PyProjectConfiguration, group: Optional["ProjectGroup"] = None):
        self._path = pyproject.path.absolute().parent
        self._pyproject = pyproject
        self._descriptor = pyproject.project.package_descriptor()
        if group:
            self.group = group  # noqa

    @property
    def config(self) -> PyProjectConfiguration:
        """
        :return: the project configuration (a.k.a., pyproject.toml)
        """
        return self._pyproject

    @cached_property
    def group(self) -> Optional["ProjectGroup"]:
        """
        :return: the project group if it belongs to such, otherwise None
        """
        from pkm.api.projects.project_group import ProjectGroup
        return ProjectGroup.of(self)

    @cached_property
    def environments_config(self) -> EnvironmentsConfiguration:
        """
        :return: the environments.toml configuration (etc/pkm/environments.toml)
        """
        return EnvironmentsConfiguration.load(self.path / ENVIRONMENT_CONFIGURATION_PATH)

    @property
    def path(self) -> Path:
        """
        :return: the path to the project root (where the pyproject.toml is located) or
                 None if this project was not loaded from a path
        """
        return self._path

    @cached_property
    def published_metadata(self) -> Optional[PackageMetadata]:
        return PackageMetadata.from_project_config(self.config.project)

    @cached_property
    def computed_metadata(self) -> PackageMetadata:
        if self.config.project.dynamic:
            dist_info = DistInfo.load(self.build_wheel(only_meta=True))
            return dist_info.load_metadata_cfg()
        return self.published_metadata

    @property
    def descriptor(self) -> PackageDescriptor:
        return self._descriptor

    def _all_dependencies(self, environment: "Environment") -> List["Dependency"]:
        prj = self.config.project

        if prj.is_dynamic('dependencies') or prj.is_dynamic('optional-dependencies'):
            return self.computed_metadata.dependencies

        return self._pyproject.project.all_dependencies

    def is_compatible_with(self, env: "Environment") -> bool:
        return self._pyproject.project.requires_python.allows_version(env.interpreter_version)

    def install_to(
            self, target: PackageInstallationTarget, user_request: Optional["Dependency"] = None,
            editable: bool = True):

        from pkm.api.distributions.wheel_distribution import WheelDistribution
        with temp_dir() as tdir, PackageInstallMonitoredOp(self.descriptor):
            wheel = self.build_wheel(tdir, editable=editable, target_env=target.env)
            distribution = WheelDistribution(self.descriptor, wheel)
            distribution.install_to(target, RequestedPackageInfo(user_request, editable))

    def update_at(self, target: "PackageInstallationTarget", editable: bool = True):
        # fast alternative for application to be installed without passing through wheels
        if self.is_containerized_application():
            target.app_containers.install(self, editable)
        else:
            super(Project, self).update_at(target, editable)

    @cached_property
    def lock(self) -> PackagesLock:
        """
        :return: the project lock, read more about it in `PackagesLock` documentation
        """
        return PackagesLock.load(self.directories.etc_pkm / 'packages-lock.toml')

    @cached_property
    def directories(self) -> "ProjectDirectories":
        """
        :return: common project directories
        """
        return ProjectDirectories.create(self._pyproject)

    def bump_version(self, particle: str, new_name: Optional[str] = None, save: bool = True) -> Version:
        """
        bump up the version of this project
        :param particle: the particle of the version to bump, can be any of: major, minor, patch, a, b, rc, name
        :param new_name: if `particle` equals to 'name' than the new name is taken from this argument
        :param save: if true then the new configuration is saved into pyproject.toml
        :return: the new version after the bump
        """

        if particle == 'name':
            if not new_name:
                raise UnsupportedOperationException("particle was 'name' but no name was provided")

            new_version = NamedVersion(new_name)
        else:
            version: Version = self.config.project.version
            if not isinstance(version, StandardVersion) or not len(version.release) == 3:
                raise UnsupportedOperationException("cannot bump version that does not follow the semver semantics")

            new_version = version.bump(particle)

        self.config.project = replace(self.config.project, version=new_version)
        if save:
            self.config.save()
        return new_version

    def remove_dependencies(self, packages: List[str]):
        """
        remove and uninstall all dependencies that are related to the given list of packages
        :param packages: the list of package names to remove
        """

        package_names_set = set(packages)
        project_dependencies = self._pyproject.project.dependencies or []
        self._pyproject.project = replace(
            self._pyproject.project,
            dependencies=[d for d in project_dependencies if d.package_name not in package_names_set])
        self._pyproject.save()

        # fix installation metadata of the project by reinstalling it (without dependencies)
        self.update_at(self.attached_environment.installation_target)
        self.attached_environment.uninstall(packages)

        self.lock.update_lock(self.attached_environment)
        self.lock.save()

    def install_with_dependencies(
            self, new_dependencies: Optional[List["Dependency"]] = None,
            optional_group: Optional[str] = None, update_existing: bool = False):
        """
        install the dependencies of this project to its assigned environments
        :param new_dependencies: if given, resolve and add these dependencies to this project and then install
        :param optional_group: if not None, installs the dependencies including the ones from the given group,
           also, mark the newly installed dependencies as optional and add them to that group
        :param update_existing: if True, will attempt to update the given `new_dependencies`,
           or all the project dependencies if no `new_dependencies` are given
        """

        deps = {d.package_name: d for d in (self._pyproject.project.dependencies or [])}
        if optional_group:
            deps.update(
                {d.package_name: d for d in (self._pyproject.project.optional_dependencies.get(optional_group) or [])})

        new_deps: Dict[str, "Dependency"] = {d.package_name: d for d in new_dependencies} if new_dependencies else {}

        # save the new dependencies to the configuration, use a wide version specifier if no such specifier is given:
        _update_dependencies(self.config, new_deps, optional_group)

        if update_existing:
            self.lock.unlock_packages(new_deps.keys()).save()

        repository = self.attached_repository
        project_dependency = self.descriptor.to_dependency()
        if optional_group:
            project_dependency = project_dependency.with_extras([optional_group])
        self.update_at(self.attached_environment.installation_target)  # should probably submit the optionals
        self.attached_environment.install(
            project_dependency, repository, packages_to_update=list(new_deps.keys()) if update_existing else None)

        # attempt to update pyproject to contain a more specific dependency version specifications
        new_deps_with_version = {}
        site_packages = self.attached_environment.site_packages
        for dep in new_deps.values():
            if not dep.version_spec.is_any():
                spec = dep.version_spec
            else:
                installed_package = site_packages.installed_package(dep.package_name)
                assert installed_package, \
                    f"package: {dep.package_name} could not be found in site packages after installation"

                installed_version = installed_package.version
                if isinstance(installed_version, StandardVersion):
                    spec = VersionRange(
                        installed_version,
                        replace(installed_version, release=(installed_version.release[0] + 1,)),
                        True, False)
                else:
                    spec = SpecificVersion(installed_version)

            new_deps_with_version[dep.package_name] = replace(dep, version_spec=spec)

        _update_dependencies(self.config, new_deps_with_version, optional_group)
        self.lock.update_lock(self.attached_environment)
        self.lock.save()

    def _reload(self):
        clear_cached_properties(self)

    @cached_property
    def attached_environment(self) -> "Environment":
        """
        :return: the virtual environment that is attached to this project
        """

        cfg = self.environments_config.attached_env
        if not cfg and self.group:
            cfg = self.group.environments_config.attached_env

        cfg = cfg or AttachedEnvironmentConfig()

        if not cfg.path and not cfg.zoo:
            env_path = self.path / ".venv"
        elif cfg.path:
            env_path = cfg.path
        else:
            env_path = cfg.zoo / self.name

        from pkm.api.environments.environment import Environment
        from pkm.api.dependencies.dependency import Dependency

        if not Environment.is_valid(env_path):
            return EnvironmentBuilder.create_matching(
                env_path, Dependency('python', self.config.project.requires_python))
        return Environment(env_path)

    @cached_property
    def attached_repository(self) -> "Repository":
        """
        :return: the repository that is attached to this project (will be used with its attached environment)
        """
        return pkm.repository_loader.load_for_project(self)

    def build_app_sdist(self, target_dir: Optional[Path] = None) -> Path:
        """
        build a containerized application source distribution from this project
        :param target_dir: the directory to put the created archive in
        :return: the path to the created archive
        """
        cnt_app_prj = Project.load(self.path)
        cnt_app_prj.config.pkm_distribution = PkmDistributionConfig(PKM_DIST_CFG_TYPE_LIB)
        cnt_app_prj.config.pkm_application = PkmApplicationConfig(True, self.config.project.dependencies, {}, [])
        cnt_app_prj.config.project = replace(cnt_app_prj.config.project, dependencies=[])
        return cnt_app_prj.build_sdist(target_dir)

    def build_sdist(self, target_dir: Optional[Path] = None) -> Path:
        """
        build a source distribution from this project
        :param target_dir: the directory to put the created archive in
        :return: the path to the created archive
        """

        if self.is_pkm_project():
            from pkm.pep517_builders.pkm_builders import build_sdist
            return build_sdist(self, target_dir)
        else:
            from pkm.pep517_builders.external_builders import build_sdist
            return build_sdist(self, target_dir)

    def build_wheel(self, target_dir: Optional[Path] = None, only_meta: bool = False, editable: bool = False,
                    target_env: Optional[Environment] = None) -> Path:
        """
        build a wheel distribution from this project
        :param target_dir: directory to put the resulted wheel in
        :param only_meta: if True, only builds the dist-info directory otherwise the whole wheel
        :param editable: if True, a wheel for editable install will be created
        :param target_env: the environment that this build should be compatible with, defaults to attached env
        :return: path to the built artifact (directory if only_meta, wheel archive otherwise)
        """
        if self.is_containerized_application() and not only_meta:
            from pkm.pep517_builders.pkm_app_builders import build_wheel
            return build_wheel(self, target_dir, editable=editable, target_env=target_env)
        elif self.is_pkm_project():
            from pkm.pep517_builders.pkm_builders import build_wheel
            return build_wheel(self, target_dir, only_meta, editable, target_env=target_env)
        else:
            from pkm.pep517_builders.external_builders import build_wheel
            return build_wheel(self, target_dir, only_meta, editable, target_env=target_env)

    def build(self, target_dir: Optional[Path] = None) -> List[Path]:
        """
        builds the project into all distributions that are required as part of its configuration
        :param target_dir: directory to put the resulted distributions in
        :return list of paths to all the distributions created
        """

        builders: List[Callable[[Path], Path]] = [self.build_sdist, self.build_wheel]
        if self.is_containerized_application():
            builders = [self.build_sdist]
        elif self.config.pkm_distribution.type == PKM_DIST_CFG_TYPE_CAPP:
            builders = [self.build_app_sdist]

        return [build(target_dir) for build in builders]

    def is_pkm_project(self) -> bool:
        """
        :return: True if this project is a pkm project, False otherwise
        """
        return self.config.build_system.build_backend == 'pkm.api.buildsys'

    def is_containerized_application(self) -> bool:
        """
        :return: true if this project represents a containerized application project
                 (the `tool.pkm.application` section exists)
        """
        return self._pyproject.pkm_application is not None

    def is_built_in_default_location(self) -> bool:
        """
        :return: True if the project default dist folder contain a build directory for the current version,
                 False otherwise
        """

        return (self.directories.dist / str(self.version)).exists()

    def publish(self, repository: Union[Repository, RepositoryPublisher], auth: Authentication,
                distributions_dir: Optional[Path] = None):
        """
        publish/register this project distributions, as found in the given `distributions_dir`
        to the given `repository`. using `auth` for authentication

        :param repository: the repository to publish to
        :param auth: authentication for this repository
        :param distributions_dir: directory containing the distributions (archives like wheels and sdists) to publish
        """

        distributions_dir = distributions_dir or (self.directories.dist / str(self.version))

        if not distributions_dir.exists():
            raise FileNotFoundError(f"{distributions_dir} does not exists")

        publisher = repository if isinstance(repository, RepositoryPublisher) else repository.publisher
        if not publisher:
            raise UnsupportedOperationException(f"the given repository ({repository.name}) is not publishable")

        metadata = PackageMetadata.from_project_config(self._pyproject.project)
        for distribution in distributions_dir.iterdir():
            if distribution.is_file():
                publisher.publish(auth, metadata, distribution)

    @classmethod
    def load(cls, path: Union[Path, str], package: Optional[PackageDescriptor] = None,
             group: Optional["ProjectGroup"] = None) -> "Project":
        path = Path(path)
        pyproject = PyProjectConfiguration.load_effective(path / 'pyproject.toml', package)
        return Project(pyproject, group=group)


@dataclass()
class ProjectDirectories:
    src_packages: List[Path]
    dist: Path
    etc_pkm: Path

    def clean_dist(self, keep_versions: Optional[List[Version]] = None):
        """
        delete the build artifacts from dist
        :param keep_versions: if given, all artifacts that relate to the versions in the given list will not be deleted
        """

        if not self.dist.exists():
            return

        keep_versions = set(str(p) for p in (keep_versions or []))

        for file in self.dist.iterdir():
            if file.is_dir():
                if file.name in keep_versions:
                    continue
                shutil.rmtree(file)
            else:
                file.unlink()

    @classmethod
    def create(cls, pyproject: PyProjectConfiguration) -> "ProjectDirectories":
        project_path = pyproject.path.parent
        packages_relative = pyproject.pkm_project.packages
        if packages_relative:
            packages = [project_path / p for p in packages_relative]
        else:
            if not (src_dir := project_path / 'src').exists():
                src_dir = project_path
            packages = [p for p in src_dir.iterdir() if p.is_dir()]

        etc_pkm = project_path / 'etc' / 'pkm'
        etc_pkm.mkdir(parents=True, exist_ok=True)
        return ProjectDirectories(packages, project_path / 'dist', etc_pkm)


def _update_dependencies(
        config: PyProjectConfiguration, new_deps: Dict[str, "Dependency"], optional_group: Optional[str]):
    save_dependencies = [d for d in (config.project.dependencies or []) if d.package_name not in new_deps]
    save_optional_dependencies = config.project.optional_dependencies

    if optional_group:
        save_optional_dependencies[optional_group] = \
            [d
             for d in save_optional_dependencies.get(optional_group, [])
             if d.package_name not in new_deps] + list(new_deps.values())
    else:
        save_dependencies = save_dependencies + list(new_deps.values())

    config.project = replace(
        config.project,
        dependencies=save_dependencies, optional_dependencies=save_optional_dependencies)
    config.save()



import os
import platform
from pathlib import Path
from typing import Literal

OS: Literal['Windows', 'Linux', 'Darwin', 'Java', ''] = platform.system()  # noqa


def is_windows() -> bool:
    """
    :return: true if running inside a windows OS
    """
    return OS == 'Windows'


def is_executable(file: Path) -> bool:
    """
    :param file: the file to test
    :return: true if the given [file] is executable (os independent)
    """
    if OS == 'Windows':
        return file.suffix == '.exe'
    return os.access(str(file), os.X_OK)



from pathlib import Path
from typing import Optional, TYPE_CHECKING

from pkm.api.distributions.distinfo import RequestedPackageInfo
from pkm.api.distributions.distribution import Distribution
from pkm.api.distributions.wheel_distribution import WheelDistribution
from pkm.api.packages.package import PackageDescriptor
from pkm.api.packages.package_installation import PackageInstallationTarget
from pkm.api.packages.package_metadata import PackageMetadata
from pkm.api.pkm import pkm

if TYPE_CHECKING:
    from pkm.api.environments.environment import Environment


class SourceDistribution(Distribution):

    def __init__(self, package: PackageDescriptor, archive: Path):
        self._package = package
        self.archive = archive

    @property
    def owner_package(self) -> PackageDescriptor:
        return self._package

    def extract_metadata(self, env: "Environment") -> PackageMetadata:
        return pkm.source_build_cache.get_or_build_meta(env, self)

    def install_to(
            self, target: PackageInstallationTarget, user_request: Optional[RequestedPackageInfo] = None):
        WheelDistribution(self.owner_package, pkm.source_build_cache.get_or_build_wheel(target, self)) \
            .install_to(target, user_request)



from collections import defaultdict
from pathlib import Path
from typing import List, Dict, Optional, Any

from pkm.api.dependencies.dependency import Dependency
from pkm.api.distributions.distribution import Distribution
from pkm.api.packages.package import Package
from pkm.api.projects.project import Project
from pkm.api.projects.project_group import ProjectGroup
from pkm.api.repositories.repository import AbstractRepository, RepositoryBuilder, Repository
from pkm.utils.strings import endswith_any

LOCAL_PACKAGES_REPOSITORY_TYPE = "local"


class PackagesDictRepository(AbstractRepository):

    def __init__(self, name: str, packages: Dict[str, List[Package]]):
        super().__init__(name)
        self._packages = packages

    def _do_match(self, dependency: Dependency) -> List[Package]:
        all_packages = self._packages.get(dependency.package_name) or []
        return [p for p in all_packages if dependency.version_spec.allows_version(p.version)]


class LocalPackagesRepositoryBuilder(RepositoryBuilder):

    def __init__(self):
        super().__init__(LOCAL_PACKAGES_REPOSITORY_TYPE)

    def build(self, name: Optional[str], packages: Optional[List[str]], **kwargs: Any) -> Repository:
        projects: List[str] = kwargs.get('projects', [])
        distributions: List[str] = kwargs.get('distributions', [])

        packages: Dict[str, List[Package]] = defaultdict(list)

        # load projects
        for project_path_str in projects:
            project_path = Path(project_path_str)
            if (project_path / 'pyproject-group.toml').exists():
                project_group = ProjectGroup.load(project_path)
                for project in project_group.project_children_recursive:
                    packages[project.name].append(project)
            else:
                project = Project.load(project_path)
                packages[project.name].append(project)

        # load distributions
        for distribution_path_str in distributions:
            distribution_path = Path(distribution_path_str)
            if distribution_path.is_dir():
                for dist in distribution_path.iterdir():
                    if endswith_any(dist.name, ('.tar.gz', '.whl')):
                        package = Distribution.package_from(dist)
                        packages[package.name].append(package)
            else:
                package = Distribution.package_from(distribution_path)
                packages[package.name].append(package)

        return PackagesDictRepository(name, packages)



from dataclasses import dataclass

from pkm.api.packages.package import PackageDescriptor
from pkm.utils.monitors import MonitoredOperation


@dataclass
class PackageInstallMonitoredOp(MonitoredOperation):
    package: PackageDescriptor



import tarfile
from io import UnsupportedOperation
from pathlib import Path
from zipfile import ZipFile

from pkm.utils.strings import endswith_any


def extract_archive(archive_path: Path, target_directory: Path):
    target_directory.mkdir(exist_ok=True, parents=True)

    if endswith_any(archive_path.name, (".zip", '.whl')):
        with ZipFile(archive_path) as z:
            z.extractall(target_directory)
    elif archive_path.name.endswith('.tar.gz'):
        with tarfile.open(archive_path) as tar:
            tar.extractall(target_directory)
    else:
        raise UnsupportedOperation(f"does not support archive {archive_path.name}")



from pathlib import Path
from typing import List

from dataclasses import replace

from pkm.utils.http.http_client import Url
from pkm.utils.parsers import SimpleParser
from pkm.api.versions.version import Version, NamedVersion, StandardVersion, UrlVersion
from pkm.api.versions.version_specifiers import VersionSpecifier, VersionRange, SpecificVersion, AnyVersion
import re

_VERSION_URL_RX = re.compile(r"((?P<repo>\w+)\+)?(?P<url>.*)")


def parse_version(version_str: str) -> Version:
    try:
        return VersionParser(version_str.lower()).read_version()
    except ValueError:
        return NamedVersion(version_str)


def parse_specifier(specifier_str: str) -> VersionSpecifier:
    return VersionParser(specifier_str.lower()).read_specifier()


_PRE_RELEASE_TYPE_NORMALIZER = {
    'a': 'a', 'b': 'b', 'rc': 'rc', 'c': 'rc', 'alpha': 'a', 'beta': 'b', 'preview': 'rc'
}


class VersionParser(SimpleParser):

    def read_version(self) -> StandardVersion:
        self.match_ws()

        wrapping_brakets = self.match("(", "")
        wrapping_quotes = self.match("'", "") or self.match("\"", "")

        self.match('v')

        epoch = None
        n = self.read_digits()
        if self.match('!'):
            epoch = int(n)
            n = self.read_digits()

        release: List[int] = [int(n)]
        while self.match('.') and self.is_not_empty() and self.peek().isdigit():
            release.append(int(self.read_digits()))

        pre_release = None
        # pre_release_type = self.peek_prev() != '.' and self.match_any(*_PRE_RELEASE_TYPE_NORMALIZER.keys())
        pre_release_type = self.match_any(*_PRE_RELEASE_TYPE_NORMALIZER.keys())
        if pre_release_type:
            self.match_any('.', '-', '_')
            pre_release_num = 0 if not self.peek().isdigit() else int(self.read_digits())
            pre_release = (_PRE_RELEASE_TYPE_NORMALIZER[pre_release_type], pre_release_num)

        post_release = None

        if (prev := self.peek_prev()).isdigit() or prev in _PRE_RELEASE_TYPE_NORMALIZER:
            self.match_any('.', '-', '_')
        if self.match_any('post', 'r', 'rev') or (self.peek_prev() == '-' and self.peek().isdigit()):
            self.match('-')
            post_release = 0 if not self.peek().isdigit() else int(self.read_digits())

        dev_release = None
        if self.peek_prev().isdigit():
            self.match_any('.', '-', '_')
        if self.match('dev'):
            dev_release = 0 if not self.peek().isdigit() else int(self.read_digits())

        local_label = None
        if self.match('+'):
            local_label = self.until(lambda i, s: s[i] not in ".-_" and not s[i].isalnum())
            local_label = local_label.translate(local_label.maketrans('-_', '..'))

        if self.peek_prev() == '.':
            self.position -= 1

        self.match_ws()

        if wrapping_quotes:
            self.match_or_err(wrapping_quotes, f"expecting closing quote: [{wrapping_quotes}]")
        if wrapping_brakets:
            self.match_or_err(')', f"expecting closing quote: [)]")
        self.match_ws()

        return StandardVersion(release=tuple(release), epoch=epoch, pre_release=pre_release, post_release=post_release,
                               dev_release=dev_release, local_label=local_label)

    # noinspection PyShadowingBuiltins
    def _read_single_specifier(self) -> VersionSpecifier:
        self.match_ws()

        if self.match('*'):
            return AnyVersion

        if self.match('@'):
            self.match_ws()
            url = self.until(lambda i, s: s[i].isspace())
            if "://" not in url:
                url = Path(url).resolve().as_uri()

            match = _VERSION_URL_RX.match(url)
            parsed_url = Url.parse(match.group('url'))
            return SpecificVersion(UrlVersion(match.group('url'), match.group('repo'), parsed_url.scheme))

        if self.match('==='):
            self.match_ws()
            version_str = self.until(lambda i, t: t[i].isspace() or t[i] == ',')
            return SpecificVersion(NamedVersion(version_str))

        exclusion_inclusion = self.match_any('==', '!=')
        if exclusion_inclusion:
            self.match_ws()
            version = self.read_version()
            if self.match('.*'):
                result = VersionRange(
                    min=version, includes_min=True,
                    max=replace(version, release=version.release[:-1] + (version.release[-1] + 1,),
                                pre_release=None, post_release=None, dev_release=None, local_label=None),
                    includes_max=False)
            else:
                result = SpecificVersion(version)

            if exclusion_inclusion == '!=':
                return result.inverse()
            return result

        if self.match('~='):
            self.match_ws()
            version = self.read_version()

            if len(version.release) == 1:
                raise ValueError(f"Illegal version specifier, ~= {version}")

            return VersionRange(
                min=version, includes_min=True,
                max=replace(version, release=version.release[:-2] + (version.release[-2] + 1, 0),
                            pre_release=None, post_release=None, dev_release=None, local_label=None),
                includes_max=False)

        comparison = self.match_any('>=', '>', '<=', '<')
        if comparison:
            self.match_ws()
            version = self.read_version()
            min, max, imin, imax = None, None, False, False

            if comparison[0] == '>':
                min, imin = version, len(comparison) > 1
            else:
                max, imax = version, len(comparison) > 1

            self.match(".*")  # version spec can end (un-needed-ly for comparison operators other than ==/!=) with .*

            return VersionRange(min=min, max=max, includes_min=imin, includes_max=imax)

        self.raise_err('unknown operator')

    def read_specifier(self):
        self.match_ws()
        paren = self.match('(')

        specifier = AnyVersion
        while self.is_not_empty():
            specifier = specifier.intersect(self._read_single_specifier())
            self.match_ws()
            if not self.match(',') and self.is_not_empty():
                if not paren or self.match(')'):
                    break
                self.raise_err("expecting comma (',')")

        return specifier



import zipfile
from pathlib import Path
from typing import Optional
from zipfile import ZipFile

from pkm.api.distributions.wheel_distribution import WheelDistribution
from pkm.api.environments.environment import Environment
from pkm.api.environments.environment_builder import EnvironmentBuilder
from pkm.api.projects.project import Project
from pkm.utils.commons import UnsupportedOperationException
from pkm.utils.files import temp_dir

CONTAINERIZED_APP_SITE_PATH = "__packages__"
CONTAINERIZED_APP_BIN_PATH = "__bin__"
CONTAINERIZED_APP_DATA_PATH = "__data__"


def build_wheel(project: "Project", target_dir: Optional[Path] = None,
                editable: bool = False,
                target_env: Optional[Environment] = None) -> Path:
    """
    build a wheel distribution from this project
    :param project: the project to build
    :param target_dir: directory to put the resulted wheel in
    :param editable: if True, a wheel for editable install will be created
    :param target_env: the environment that this build should be compatible with, defaults to the project's attached env
    :return: path to the built artifact (directory if only_meta, wheel archive otherwise)
    """

    print("building app wheel")

    target_dir = target_dir or (project.directories.dist / str(project.version))
    interpreter_path = target_env.interpreter_path if target_env else project.attached_environment.interpreter_path

    if not (app_config := project.config.pkm_application) or not app_config.containerized:
        raise UnsupportedOperationException("not application or non containerized")

    with temp_dir() as tdir:
        tenv = EnvironmentBuilder.create(tdir / "env", interpreter_path)
        target = tenv.installation_target

        target.app_containers.install(project, editable=editable)

        # build the actual wheel
        wheel_path = target_dir / WheelDistribution.expected_wheel_file_name(project)
        target_dir.mkdir(parents=True, exist_ok=True)
        with ZipFile(wheel_path, 'w', compression=zipfile.ZIP_DEFLATED) as wheel:
            build_dir = Path(target.purelib)
            for file in build_dir.rglob('*'):
                wheel.write(file, file.relative_to(build_dir))

        return wheel_path



import hashlib
import shutil
from pathlib import Path
from typing import Any

import pkm.pep517_builders.external_builders as ext_build
from pkm.api.distributions.source_distribution import SourceDistribution
from pkm.api.distributions.wheel_distribution import WheelDistribution
from pkm.api.environments.environment import Environment
from pkm.api.packages.package import PackageDescriptor
from pkm.api.packages.package_installation import PackageInstallationTarget
from pkm.api.packages.package_metadata import PackageMetadata
from pkm.api.projects.project import Project
from pkm.pep517_builders.external_builders import BuildError
from pkm.utils.archives import extract_archive
from pkm.utils.files import temp_dir
from pkm.utils.hashes import stream
from pkm.utils.iterators import single_or, single_or_raise

_normalize = PackageDescriptor.normalize_src_package_name


class SourceBuildCache:
    def __init__(self, workspace: Path):
        self.workspace = workspace
        workspace.mkdir(parents=True, exist_ok=True)

    def _version_dir(self, package: PackageDescriptor) -> Path:
        return self.workspace / package.name / _normalize(str(package.version))

    def get_or_build_wheel(self, target: PackageInstallationTarget, dist: SourceDistribution) -> Path:
        return self._get_or_build(target, dist, 'wheel')

    def get_or_build_meta(self, env: Environment, dist: SourceDistribution) -> PackageMetadata:
        return self._get_or_build(env.installation_target, dist, 'metadata')

    def _get_or_build(self, target: PackageInstallationTarget, dist: SourceDistribution, artifact: str) -> Any:

        base_cache_dir = self.workspace / _normalize(dist.owner_package.name) / \
                         _normalize(str(dist.owner_package.version)) / \
                         target.env.markers_hash

        dist_hash = hashlib.blake2s()
        stream(dist_hash, dist.archive)
        dist_hash_value = dist_hash.hexdigest()
        metadata = artifact == 'metadata'
        cache_dir = base_cache_dir / dist_hash_value
        metadata_file = cache_dir / "METADATA"

        if metadata_file.exists():
            if metadata:
                return PackageMetadata.load(metadata_file)
            if wheel := single_or(cache_dir.glob("*.whl"), None):
                return wheel

        if base_cache_dir.exists():
            shutil.rmtree(base_cache_dir)

        with temp_dir() as tdir:
            extract_archive(dist.archive, tdir)
            sdir = single_or_raise(tdir.iterdir())
            project = Project.load(sdir, dist.owner_package)
            odir = tdir / 'output'

            try:
                output = ext_build.build_wheel(
                    project, odir, only_meta=metadata, editable=False, target_env=target.env)
            except BuildError as e:
                if not metadata or not e.missing_hook:
                    raise
                output = ext_build.build_wheel(
                    project, odir, only_meta=False, target_env=target.env)

            if output.is_dir():  # metadata
                wheel_metadata_path = output / 'METADATA'
                if not wheel_metadata_path.exists():
                    raise BuildError("build backend did not create a wheel METADATA file")

                cache_dir.mkdir(parents=True, exist_ok=True)
                shutil.copy(wheel_metadata_path, metadata_file)
                return PackageMetadata.load(wheel_metadata_path)
            else:
                if not metadata_file.exists():
                    metadata = WheelDistribution(dist.owner_package, output).extract_metadata()
                    metadata.save_to(metadata_file)

                cached_output = cache_dir / output.name
                cache_dir.mkdir(parents=True, exist_ok=True)
                shutil.move(output, cached_output)
                return cached_output



from abc import ABC, ABCMeta
from typing import Type, TypeVar, Union, Callable, Tuple, Dict

from pkm.utils.commons import UnsupportedOperationException

_T = TypeVar("_T")
_U = TypeVar("_U")

_ABSTRACT_METHODS_KEY = '__abstractmethods__'


def delegate(type_: Type[_T], attr: str) -> Callable[[Type[_U]], Type[Union[_T, _U]]]:
    return delegate_all((type_, attr))


def delegate_all(*reqs: Tuple[Type[_T], str]) -> Callable[[Type[_U]], Type[Union[_T, _U]]]:
    method_to_attr: Dict[str, str] = {}

    for type_, attr in reqs:
        if not hasattr(type_, '__dict__'):
            raise UnsupportedOperationException(
                f"cannot create delegations using classes that has no __dict__ ({type_})")

        method_to_attr.update({
            method: attr
            for method in type_.__dict__[_ABSTRACT_METHODS_KEY]
        })

    def delegate_(cls: Type[_U]) -> Type[Union[_T, _U]]:

        if not hasattr(cls, '__dict__'):
            raise UnsupportedOperationException("cannot create delegations for classes that has no __dict__")

        if _ABSTRACT_METHODS_KEY not in cls.__dict__:
            # no delegation is needed - all methods are already implemented..
            return cls

        # first copy the previous dict but without the abstract methods, we are going to implement them in a second
        new_dict = {
            attr_k: attr_v
            for attr_k, attr_v in cls.__dict__.items()
            if attr_k not in (_ABSTRACT_METHODS_KEY, '_abc_impl')
        }

        # now 'implement' the abstract methods using the type requirements
        abstract_methods = []
        for method in cls.__dict__[_ABSTRACT_METHODS_KEY]:
            if attr_m := method_to_attr.get(method):
                def _create(attr, method): # noqa
                    def do_delegation(self, *args, **kwargs):
                        return getattr(getattr(self, attr), method)(*args, **kwargs)

                    new_dict[method] = do_delegation

                _create(attr_m, method)
            else:
                abstract_methods.append(method)

        if abstract_methods:
            # we are not creating a concrete class so we should use abcmeta to create it
            return ABCMeta(cls.__name__, cls.__bases__, new_dict)

        # we are creating a concrete class
        return type(cls.__name__, tuple(b for b in cls.__bases__ if b is not ABC), new_dict)

    return delegate_



from abc import ABC, abstractmethod
from dataclasses import dataclass, replace
from typing import Tuple, Optional, Literal

from pkm.utils.commons import UnsupportedOperationException


class Version(ABC):
    @abstractmethod
    def is_pre_or_dev_release(self) -> bool:
        ...

    @abstractmethod
    def is_post_release(self) -> bool:
        ...

    @abstractmethod
    def is_local(self) -> bool:
        ...

    def without_patch(self):
        return self

    def without_local(self) -> "Version":
        return self

    def __lt__(self, other: "Version") -> bool:
        return _less(self, other)

    @abstractmethod
    def __eq__(self, other) -> bool:
        ...

    @classmethod
    def parse(cls, txt: str) -> "Version":
        from pkm.api.versions.version_parser import parse_version
        return parse_version(txt)


@dataclass(frozen=True)
class UrlVersion(Version):
    url: str
    _protocol: str
    _schema: str

    @property
    def protocol(self) -> str:
        return self._protocol or self._schema

    def is_pre_or_dev_release(self) -> bool:
        return False

    def is_post_release(self) -> bool:
        return False

    def is_local(self) -> bool:
        return False

    def __eq__(self, other) -> bool:
        return isinstance(other, UrlVersion) and self.protocol == other.protocol and self.url == other.url

    def __str__(self):
        if self._protocol:
            return self._protocol + "+" + self.url
        return self.url

    def __repr__(self):
        return str(self)


@dataclass(frozen=True)
class NamedVersion(Version):
    name: str

    def is_pre_or_dev_release(self) -> bool:
        return False

    def is_post_release(self) -> bool:
        return False

    def is_local(self) -> bool:
        return False

    def __eq__(self, other) -> bool:
        return isinstance(other, NamedVersion) and self.name == other.name

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"NamedVersion({str(self)})"


@dataclass(frozen=True, repr=False)
class StandardVersion(Version):
    release: Tuple[int, ...]
    epoch: Optional[int] = None
    pre_release: Optional[Tuple[Literal['a', 'b', 'rc'], int]] = None
    post_release: Optional[int] = None
    dev_release: Optional[int] = None
    local_label: Optional[str] = None

    def without_local(self) -> "StandardVersion":
        if self.is_local():
            return replace(self, local_label=None)
        return self

    def without_patch(self):
        if len(self.release) > 2:
            return replace(self, release=self.release[:2])
        return self

    def is_pre_or_dev_release(self) -> bool:
        return self.pre_release is not None or self.dev_release is not None

    def is_post_release(self) -> bool:
        return self.post_release is not None

    def is_local(self) -> bool:
        return self.local_label is not None

    def bump(self, particle: str) -> Version:
        """
        bump up this version
        :param particle: the particle of the version to bump, can be any of: major, minor, patch, a, b, rc
        :return: the new version after the bump
        """

        version: Version = self
        if not isinstance(version, StandardVersion) or not len(version.release) == 3:
            raise UnsupportedOperationException("cannot bump version that does not follow the semver semantics")

        def update_release(v: StandardVersion, new_release: Tuple[int, ...], ) -> Version:
            return replace(v, release=new_release,
                           pre_release=None, post_release=None, dev_release=None, local_label=None)

        def bump_release(v: StandardVersion, p: str) -> Optional[Version]:
            release = v.release
            if p == 'major':
                return update_release(v, (release[0] + 1, 0, 0))
            elif p == 'minor':
                return update_release(v, (release[0], release[1] + 1, 0))
            elif p == 'patch':
                return update_release(v, (release[0], release[1], release[2] + 1))
            else:
                return None

        def bump_pre_release(v: StandardVersion, p: str) -> Optional[Version]:
            pre_release = v.pre_release
            if pre_release:
                if pre_release[0] > p:
                    v = bump_release(v, 'patch')
                    pre_release = None
                elif pre_release[0] < p:
                    pre_release = (p, 0)
                else:
                    pre_release = (p, pre_release[1] + 1)
            else:
                pre_release = (p, 0)

            return replace(v, pre_release=pre_release)

        new_version = bump_release(version, particle) or bump_pre_release(version, particle)
        if not new_version:
            raise ValueError(f"unknown version particle: {particle}")

        return new_version

    def __str__(self):
        vstring = ''
        if self.epoch is not None:
            vstring += f'{self.epoch}!'
        vstring += '.'.join(str(n) for n in self.release)
        if self.pre_release is not None:
            vstring += ''.join(str(it) for it in self.pre_release)
        if self.post_release is not None:
            vstring += f'.post{self.post_release}'
        if self.dev_release is not None:
            vstring += f'.dev{self.dev_release}'
        if self.local_label is not None:
            vstring += f'+{self.local_label}'

        return vstring

    def __repr__(self):
        return str(self)

    def __le__(self, other):
        return other == self or self < other

    def __eq__(self, other):
        if not isinstance(other, StandardVersion):
            return False

        v1, v2 = StandardVersion.normalized(self, other)
        return (v1.epoch or 0) == (other.epoch or 0) \
               and v1.release == v2.release \
               and v1.pre_release == v2.pre_release \
               and v1.post_release == v2.post_release \
               and v1.dev_release == v2.dev_release \
               and v1.local_label == v2.local_label  # noqa

    @staticmethod
    def normalized(v1: "StandardVersion", v2: "StandardVersion") -> Tuple["StandardVersion", "StandardVersion"]:
        v1rel, v2rel = v1.release, v2.release  # noqa
        rlen = max(len(v1rel), len(v2rel))  # noqa

        if len(v1rel) != rlen:
            v1rel += (0,) * (rlen - len(v1rel))
        if len(v2rel) != rlen:
            v2rel += (0,) * (rlen - len(v2rel))

        return ((v1 if v1rel == v1.release else replace(v1, release=v1rel)),
                (v2 if v2rel == v2.release else replace(v2, release=v2rel)))

    @classmethod
    def parse(cls, txt: str) -> "StandardVersion":
        from pkm.api.versions.version_parser import VersionParser
        return VersionParser(txt.lower()).read_version()


def _less(v1: Version, v2: Version) -> bool:
    swap = False

    # order: [urls] [named-versions] [standard-versions]

    if isinstance(v1, UrlVersion) or (swap := isinstance(v2, UrlVersion)):
        if swap: return not _less(v2, v1)  # noqa

        if isinstance(v2, UrlVersion):
            return str(v1) < str(v2)

        return True

    if isinstance(v1, NamedVersion) or (swap := isinstance(v2, NamedVersion)):
        if swap: return not _less(v2, v1)  # noqa

        if isinstance(v2, NamedVersion):
            return v1.name < v2.name

        return True

    # must be standard version
    v1: StandardVersion
    v2: StandardVersion

    if (v1.epoch or 0) != (v2.epoch or 0):
        return (v1.epoch or 0) < (v2.epoch or 0)

    v1, v2 = StandardVersion.normalized(v1, v2)
    for s, o in zip(v1.release, v2.release):
        if s == o:
            continue
        return s < o

    if v1.dev_release != v2.dev_release:
        if (v1.dev_release is None) ^ (v2.dev_release is None):
            return v2.dev_release is None

        if v1.dev_release != v2.dev_release:
            return v1.dev_release < v2.dev_release

    if v1.pre_release != v2.pre_release:
        if (v1.pre_release is None) ^ (v2.pre_release is None):
            return v2.pre_release is None

        for s, o in zip(v1.pre_release, v2.pre_release):
            if s == o: continue  # noqa
            return s < o

    if v1.post_release != v2.post_release:
        if (v1.post_release is None) ^ (v2.post_release is None):
            return v1.post_release is None

        s, o = v1.post_release, v2.post_release
        if s != o:
            return s < o

    my_local_segments = (v1.local_label or "").split('.')
    other_local_segments = (v2.local_label or "").split('.')

    for m, o in zip(my_local_segments, other_local_segments):
        if m == o:
            continue

        m_num, o_num = m.isnumeric(), o.isnumeric()
        if m_num and o_num:
            return int(m_num) < int(o_num)
        if m_num or o_num:
            return o_num

        return m < o

    return len(my_local_segments) < len(other_local_segments)



"""
an implementation of pep517 & pep660 build system
"""

from pathlib import Path

from pkm.api.projects.project import Project


# noinspection PyUnusedLocal
def build_wheel(wheel_directory: str, config_settings=None, metadata_directory=None):
    return Project.load(Path(".")).build_wheel(Path(wheel_directory)).name


# noinspection PyUnusedLocal
def build_sdist(sdist_directory: str, config_settings=None):
    return Project.load(Path(".")).build_sdist(Path(sdist_directory)).name


# noinspection PyUnusedLocal
def get_requires_for_build_wheel(config_settings=None):
    return []


# noinspection PyUnusedLocal
def prepare_metadata_for_build_wheel(metadata_directory: str, config_settings=None):
    return Project.load(Path(".")).build_wheel(Path(metadata_directory), only_meta=True).name


# noinspection PyUnusedLocal
def get_requires_for_build_sdist(config_settings=None):
    return []


# noinspection PyUnusedLocal
def build_editable(wheel_directory, config_settings=None, metadata_directory=None):
    return Project.load(Path(".")).build_wheel(Path(wheel_directory), editable=True).name


# noinspection PyUnusedLocal
def get_requires_for_build_editable(config_settings=None):
    return []


# noinspection PyUnusedLocal
def prepare_metadata_for_build_editable(metadata_directory, config_settings=None):
    return Project.load(Path(".")).build_wheel(Path(metadata_directory), only_meta=True, editable=True).name



import shutil
from pathlib import Path
from typing import List, Optional, cast, TYPE_CHECKING

from pkm.api.dependencies.dependency import Dependency
from pkm.api.packages.package import Package, PackageDescriptor
from pkm.api.packages.package_installation import PackageInstallationTarget
from pkm.api.projects.project import Project
from pkm.api.projects.project_group import ProjectGroup
from pkm.api.repositories.repository import AbstractRepository
from pkm.api.versions.version import NamedVersion
from pkm.utils.commons import NoSuchElementException
from pkm.utils.http.http_client import Url
from pkm.utils.iterators import single_or_raise
from pkm.utils.processes import monitored_run
from pkm.utils.properties import cached_property

if TYPE_CHECKING:
    from pkm.api.environments.environment import Environment


# noinspection PyMethodMayBeStatic
class _Git:
    def __init__(self):
        self._git_cmd = shutil.which('git')
        if not self._git_cmd:
            raise NoSuchElementException("could not find the git command in your path")

    def is_valid_git_dir(self, dir_: Path) -> bool:
        return (dir_ / '.git').exists()

    def _is_detached(self, dir_: Path):
        """
        :param dir_: the cloned repository directory to look at
        :return: True if the repository directory is in a detached head state, False if it in a branch
        """
        head = (dir_ / '.git/HEAD').read_text()
        return not (dir_ / '.git/heads' / head).exists()

    def update(self, package_name: str, repo_dir: Path):
        if self._is_detached(repo_dir):
            return

        monitored_run(
            f'fetch {package_name}', [self._git_cmd, 'pull'], cwd=str(repo_dir)).check_returncode()

    def clone(self, package_name: str, repository_url: str, branch_or_commit: Optional[str], target_dir: Path):
        # TODO: the monitoring should probably run as part of a larger "match" monitored operation

        try:
            monitored_run(
                f'fetch {package_name}', [self._git_cmd, 'clone', repository_url, str(target_dir)]).check_returncode()

            if branch_or_commit:
                monitored_run(
                    f'fetch {package_name}', [self._git_cmd, 'checkout', branch_or_commit],
                    cwd=str(target_dir)).check_returncode()
        except Exception:
            shutil.rmtree(target_dir, ignore_errors=True)
            raise


class GitRepository(AbstractRepository):

    def __init__(self, workspace: Path):
        super().__init__("git")
        self._workspace = workspace

    @cached_property
    def _git_client(self) -> _Git:
        return _Git()

    def _do_match(self, dependency: Dependency) -> List[Package]:

        if not (version := dependency.version_spec.specific_url()) or version.protocol != 'git':
            return []

        parts = version.url.split("@")
        url = Url.parse(parts[0])
        if (sz := len(parts)) == 1:
            branch = None
        elif sz == 2:
            branch = parts[1]
        else:
            raise ValueError(f"malformed git url: {url}")

        target_dir = self._workspace / dependency.package_name / url.host / url.path.lstrip('/')
        if branch:
            target_dir = target_dir / branch

        if target_dir.exists():
            self._git_client.update(dependency.package_name, target_dir)
        else:
            self._git_client.clone(dependency.package_name, str(url), branch, target_dir)

        desc = PackageDescriptor(dependency.package_name, version)

        if ProjectGroup.is_valid(target_dir):
            project = single_or_raise(p for p in ProjectGroup.load(target_dir).project_children_recursive
                                      if p.name == dependency.package_name)
            project.bump_version('name', cast(NamedVersion, desc.version).name, save=False)
        else:
            project = Project.load(target_dir, package=desc)

        return [_GitPackageWrapper(project)]


class _GitPackageWrapper(Package):

    def __init__(self, project: Project):
        self._project = project

    @property
    def descriptor(self) -> PackageDescriptor:
        return self._project.descriptor

    def _all_dependencies(self, environment: "Environment") -> List["Dependency"]:
        return self._project.dependencies(environment)

    def is_compatible_with(self, env: "Environment") -> bool:
        return self._project.is_compatible_with(env)

    def install_to(
            self, target: "PackageInstallationTarget", user_request: Optional["Dependency"] = None,
            editable: bool = False):
        return self._project.install_to(target, user_request, editable=editable)



from dataclasses import dataclass
from pathlib import Path

from pkm.utils.monitors import MonitoredOperation, MonitoredEvent


@dataclass
class FetchResourceMonitoredOp(MonitoredOperation):
    resource_name: str
    resource_url: str


class FetchResourceCacheHitEvent(MonitoredEvent):
    ...


@dataclass
class FetchResourceDownloadStartEvent(MonitoredEvent):
    file_size: int
    store_path: Path



from html.parser import HTMLParser
from pathlib import Path
from typing import List, Union, Tuple, Dict, Optional, Callable, Any

from pkm.api.dependencies.dependency import Dependency
from pkm.api.packages.package import Package, PackageDescriptor
from pkm.api.packages.standard_package import PackageArtifact, AbstractPackage
from pkm.api.repositories.repository import Repository, RepositoryBuilder, AbstractRepository
from pkm.api.versions.version import Version
from pkm.api.versions.version_specifiers import VersionSpecifier
from pkm.utils.http.cache_directive import CacheDirective
from pkm.utils.http.http_client import HttpClient, Url
from pkm.utils.iterators import groupby
from pkm.utils.strings import endswith_any, without_suffix


class SimpleRepository(AbstractRepository):
    """
    implementation of pep503 simple repository
    """

    def __init__(self, name: str, http_client: HttpClient, url: str):
        super().__init__(name)
        self._http_client = http_client
        self._url = url
        self._base_url = Url.parse(url).connection_part()
        self._packages: Dict[str, Dict[str, Package]] = {}  # name -> version -> package

    def _do_match(self, dependency: Dependency) -> List[Package]:
        # monitor.on_dependency_match(dependency)
        if not (version_to_package := self._packages.get(dependency.package_name)):
            data = self._http_client.fetch_resource(f"{self._url}/{dependency.package_name}").data
            extractor = _HtmlArtifactsExtractor(self._base_url)
            extractor.feed(data.read_text())

            all_artifacts = extractor.artifacts
            grouped_by_version: Dict[str, List[PackageArtifact]] = groupby(
                all_artifacts, lambda a: _extract_version(a.file_name))

            version_to_package = {
                version_str: _SimplePackage(
                    self._http_client,
                    PackageDescriptor(dependency.package_name, Version.parse(version_str)),
                    version_artifacts)

                for version_str, version_artifacts in grouped_by_version.items()
            }
            self._packages[dependency.package_name] = version_to_package

        return [p for p in version_to_package.values() if dependency.version_spec.allows_version(p.version)]


_DISTRIBUTION_EXTENSIONS = (".whl", ".tar.gz", ".zip")


def _extract_version(filename: str) -> str:
    result = without_suffix(filename, suffix := endswith_any(filename, _DISTRIBUTION_EXTENSIONS))
    if suffix == '.whl':
        return result.split("-")[1]
    else:
        return result.split("-")[-1]


class _HtmlArtifactsExtractor(HTMLParser):

    def __init__(self, base_url: str, *, convert_charrefs: bool = ...) -> None:
        super().__init__(convert_charrefs=convert_charrefs)
        self._text_handler: Optional[Callable[[str], None]] = None
        self.artifacts: List[PackageArtifact] = []
        self._base_url = base_url

    def handle_data(self, data: str) -> None:
        if self._text_handler:
            self._text_handler(data)

    def handle_endtag(self, tag: str) -> None:
        if self._text_handler:
            self._text_handler('')

    def handle_starttag(self, tag: str, attrs: List[Tuple[str, Union[str, None]]]) -> None:
        self._text_handler = \
            lambda txt: self.handle_element(tag, {kv[0]: kv[1] for kv in attrs if kv[1] is not None}, txt)

    def handle_element(self, tag: str, attrs: Dict[str, str], text: str):
        self._text_handler = None
        if tag == 'a' and 'href' in attrs:
            requires_python = None
            if requires_python_str := attrs.get('data-requires-python'):
                requires_python = VersionSpecifier.parse(requires_python_str)

            url = attrs.get('href')
            if not url.startswith('http'):
                url = f"{self._base_url}/{url.lstrip('/')}"

            if endswith_any(text, _DISTRIBUTION_EXTENSIONS):
                self.artifacts.append(PackageArtifact(text, requires_python, {'url': url}))


class _SimplePackage(AbstractPackage):
    def __init__(self, http_client: HttpClient, descriptor: PackageDescriptor,
                 artifacts: List[PackageArtifact]):
        super().__init__(descriptor, artifacts)
        self._http_client = http_client

    def _retrieve_artifact(self, artifact: PackageArtifact) -> Path:
        return self._http_client.fetch_resource(
            artifact.other_info['url'], CacheDirective.allways(),
            resource_name=f"{self.name} {self.version}"
        ).data


class SimpleRepositoryBuilder(RepositoryBuilder):
    def __init__(self, http_client: HttpClient):
        super().__init__("simple")
        self._http_client = http_client

    def build(self, name: Optional[str], packages: Optional[List[str]], **kwargs: Any) -> Repository:
        if not (url := kwargs.get('url')):
            raise KeyError("url field is required to build 'simple' repository")
        return SimpleRepository(name or url, self._http_client, str(url).rstrip('/'))



import os
import shutil
import sys
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from pkm.utils.http.http_client import HttpClient
from pkm.utils.properties import cached_property, clear_cached_properties

if TYPE_CHECKING:
    from pkm.repositories.local_pythons_repository import InstalledPythonsRepository
    from pkm.api.repositories.repository import Repository
    from pkm.api.repositories.repository_loader import RepositoryLoader
    from pkm.distributions.source_build_cache import SourceBuildCache

ENV_PKM_HOME = "PKM_HOME"


@dataclass
class _PkmRepositories:
    pypi: "Repository"
    main: "Repository"
    installed_pythons: "InstalledPythonsRepository"


class _Pkm:
    repositories: _PkmRepositories

    def __init__(self):
        self.workspace = workspace = os.environ.get(ENV_PKM_HOME) or _default_home_directory()
        workspace.mkdir(exist_ok=True, parents=True)
        self.httpclient = HttpClient(workspace / 'resources/http')
        self.threads = ThreadPoolExecutor()

    @cached_property
    def source_build_cache(self) -> "SourceBuildCache":
        from pkm.distributions.source_build_cache import SourceBuildCache
        return SourceBuildCache(self.workspace / 'build-cache')

    @cached_property
    def repository_loader(self) -> "RepositoryLoader":
        from pkm.api.repositories.repository_loader import RepositoryLoader, REPOSITORIES_CONFIGURATION_PATH
        return RepositoryLoader(
            self.workspace / REPOSITORIES_CONFIGURATION_PATH,
            self.httpclient,
            self.workspace / 'repos')

    @cached_property
    def repositories(self) -> _PkmRepositories:
        from pkm.repositories.local_pythons_repository import InstalledPythonsRepository

        return _PkmRepositories(
            self.repository_loader.pypi,
            self.repository_loader.main,
            InstalledPythonsRepository(),
        )

    def clean_cache(self):
        shutil.rmtree(self.repository_loader.workspace, ignore_errors=True)
        shutil.rmtree(self.source_build_cache.workspace, ignore_errors=True)
        shutil.rmtree(self.httpclient.workspace, ignore_errors=True)

        clear_cached_properties(self)


# the methods used for finding the default data directory were adapted from the appdirs library

def _get_win_folder() -> str:
    import ctypes

    csidl_const = 28  # "CSIDL_LOCAL_APPDATA"
    buf = ctypes.create_unicode_buffer(1024)
    ctypes.windll.shell32.SHGetFolderPathW(None, csidl_const, None, 0, buf)

    # Downgrade to short path name if have highbit chars. See
    # <http://bugs.activestate.com/show_bug.cgi?id=85099>.
    if any(ord(c) > 255 for c in buf):
        buf2 = ctypes.create_unicode_buffer(1024)
        if ctypes.windll.kernel32.GetShortPathNameW(buf.value, buf2, 1024):
            buf = buf2

    return buf.value


def _default_home_directory():
    system = sys.platform
    if system == "win32":
        path = Path(_get_win_folder())
    elif system == 'darwin':
        path = Path('~/Library/Application Support/')
    else:
        path = Path(os.getenv('XDG_DATA_HOME', "~/.local/share"))

    return (path / 'pkm').expanduser().resolve()


pkm = _Pkm()



import shutil
import tarfile
import zipfile
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from shutil import ignore_patterns
from typing import Optional, ContextManager
from zipfile import ZipFile

from pkm.api.distributions.distinfo import WheelFileConfiguration, DistInfo
from pkm.api.distributions.pth_link import PthLink
from pkm.api.distributions.wheel_distribution import WheelDistribution
from pkm.api.environments.environment import Environment
from pkm.api.packages.package_metadata import PackageMetadata
from pkm.api.projects.project import ProjectDirectories, Project
from pkm.api.projects.pyproject_configuration import ProjectConfig, PyProjectConfiguration
from pkm.utils.files import temp_dir
from pkm.utils.iterators import distinct


def build_sdist(project: Project, target_dir: Optional[Path] = None) -> Path:
    """
    build a source distribution from this project
    :param project: the project to build
    :param target_dir: the directory to put the created archive in
    :return: the path to the created archive
    """

    # with monitor.on_build(project.descriptor, 'sdist') as build_monitor:
    target_dir = target_dir or (project.directories.dist / str(project.version))
    target_dir.mkdir(parents=True, exist_ok=True)

    with _build_context(project) as bc:
        sdist_path = target_dir / bc.sdist_file_name()
        data_dir = bc.build_dir / sdist_path.name[:-len('.tar.gz')]
        data_dir.mkdir()

        dist_info_path = bc.build_dir / 'build.dist-info'
        bc.build_dist_info(dist_info_path)
        shutil.copy(dist_info_path / "METADATA", data_dir / "PKG-INFO")
        bc.pyproject.save_to(data_dir / 'pyproject.toml')

        if bc.pyproject.pkm_project.packages:
            bc.copy_sources(data_dir)
        else:
            bc.copy_sources(data_dir / 'src')

        with tarfile.open(sdist_path, 'w:gz', format=tarfile.PAX_FORMAT) as sdist:
            for file in data_dir.glob('*'):
                sdist.add(file, file.relative_to(bc.build_dir))

    return sdist_path


def build_wheel(project: Project, target_dir: Optional[Path] = None, only_meta: bool = False,
                editable: bool = False, target_env: Optional[Environment] = None) -> Path:
    """
    build a wheel distribution from this project
    :param project: the project to build
    :param target_dir: directory to put the resulted wheel in
    :param only_meta: if True, only builds the dist-info directory otherwise the whole wheel
    :param editable: if True, a wheel for editable install will be created
    :param target_env: the environment that this build should be compatible with, defaults to the project's attached env
    :return: path to the built artifact (directory if only_meta, wheel archive otherwise)
    """

    target_dir = target_dir or (project.directories.dist / str(project.version))

    if target_env is not None and \
            not (req := project.config.project.requires_python).allows_version(target_env.interpreter_version):
        raise ValueError(f"project: '{project.name}' source does not support interpreter version in environment "
                         f"(requires: {req}, got {target_env.interpreter_version})")

    with _build_context(project) as bc:
        if only_meta:
            dist_info_path = target_dir / bc.dist_info_dir_name()
            bc.build_dist_info(dist_info_path)
            return dist_info_path

        dist_info_path = bc.build_dir / bc.dist_info_dir_name()
        dist_info = bc.build_dist_info(dist_info_path)
        bc.copy_sources(bc.build_dir, editable)
        records_file = dist_info.load_record_cfg()
        records_file.sign_recursive(bc.build_dir)
        records_file.save()

        wheel_path = target_dir / bc.wheel_file_name()
        target_dir.mkdir(parents=True, exist_ok=True)
        with ZipFile(wheel_path, 'w', compression=zipfile.ZIP_DEFLATED) as wheel:
            for file in bc.build_dir.rglob('*'):
                wheel.write(file, file.relative_to(bc.build_dir))

    return wheel_path


@contextmanager
def _build_context(project: Project) -> ContextManager["_BuildContext"]:
    project_cfg: ProjectConfig = project.config.project

    project_name_underscores = project_cfg.name.replace('-', '_')

    with temp_dir() as build_dir:
        yield _BuildContext(project, build_dir, project_name_underscores)


@dataclass
class _BuildContext:
    project: Project
    build_dir: Path
    project_name_underscore: str

    @property
    def pyproject(self) -> PyProjectConfiguration:
        return self.project.config

    def _project_and_version_file_prefix(self):
        return f"{self.project_name_underscore}-{self.pyproject.project.version}"

    def wheel_file_name(self) -> str:
        return WheelDistribution.expected_wheel_file_name(self.project)

    def sdist_file_name(self) -> str:
        return f"{self._project_and_version_file_prefix()}.tar.gz"

    def dist_info_dir_name(self) -> str:
        return f'{self._project_and_version_file_prefix()}.dist-info'

    def build_dist_info(self, dst: Path) -> DistInfo:
        di = DistInfo.load(dst)

        dst.mkdir(exist_ok=True, parents=True)
        project_config: ProjectConfig = self.pyproject.project

        PackageMetadata.from_project_config(project_config).save_to(di.metadata_path())
        di.license_path().write_text(
            project_config.license_content())

        # TODO: probably later we will want to add the version of pkm in the generator..
        WheelFileConfiguration.create(generator="pkm", purelib=True).save_to(di.wheel_path())

        entrypoints = di.load_entrypoints_cfg()
        entrypoints.entrypoints = [e for entries in self.pyproject.project.entry_points.values() for e in entries]
        entrypoints.save()

        return di

    def copy_sources(self, dst: Path, link_only: bool = False):
        dirs = ProjectDirectories.create(self.pyproject)

        if link_only:
            PthLink(
                dst / f"{self._project_and_version_file_prefix()}.pth",
                links=list(distinct(p.absolute().parent for p in dirs.src_packages))
            ).save()
            return

        for package_dir in dirs.src_packages:
            destination = dst / package_dir.name
            if package_dir.exists():
                shutil.copytree(package_dir, destination, ignore=ignore_patterns('__pycache__'))
            else:
                raise FileNotFoundError(f"the package {package_dir}, which is specified in pyproject.toml"
                                        " has no corresponding directory in project")



from __future__ import annotations

from dataclasses import dataclass, replace
from typing import List, Dict, Optional, TYPE_CHECKING

from pkm.api.dependencies.dependency import Dependency
from pkm.api.packages.package import PackageDescriptor, Package
from pkm.api.versions.version import Version, UrlVersion
from pkm.api.versions.version_specifiers import SpecificVersion
from pkm.pep517_builders.external_builders import BuildError
from pkm.resolution.pubgrub import Problem, MalformedPackageException, Term, Solver
from pkm.utils.dicts import get_or_put
from pkm.utils.promises import Promise
from pkm.utils.sequences import single_or_raise, oseq_hash

if TYPE_CHECKING:
    from pkm.api.environments.environment import Environment
    from pkm.api.repositories.repository import Repository


def resolve_dependencies(root: Dependency, env: "Environment", repo: "Repository",
                         dependency_overrides: Optional[Dict[str, Dependency]] = None) -> List[Package]:
    """
    transform the give input into pubgrub's dependency resolution problem, use pubgrub to
     solve the transformed problem and transform its output to list of packages
    :param root: the root package to resolve dependencies for
    :param env: the environment that all dependencies should be compatible with
    :param repo: repository to locate dependencies
    :param dependency_overrides: package to dependency mapping that is used to override some of the dependencies
           e.g., when a dependency: x==v found (anywhere in the resolution graph) and `dependency_overrides` contains
           {'x':'>=u'} then the original dependency will be replaced with the new one.
    :return: the list of packages that are required to be installed so that the `root` dependency works correctly
    """

    problem = _PkmPackageInstallationProblem(env, repo, root, dependency_overrides)
    solver = Solver(problem, _Pkg.of(root))
    solution: Dict[_Pkg, Version] = solver.solve()

    result: List[Package] = []

    for pkg, version in solution.items():
        if pkg.extras:
            continue

        result.append(problem.opened_packages[PackageDescriptor(pkg.name, version)])

    return result


class _PkmPackageInstallationProblem(Problem):

    def __init__(self, env: "Environment", repo: "Repository", root: Dependency,
                 dependency_overrides: Optional[Dict[str, Dependency]] = None):
        self._env = env
        self._repo = repo
        self._root = root
        self._dependency_overrides = dependency_overrides or {}

        from pkm.api.pkm import pkm
        self._threads = pkm.threads

        self.opened_packages: Dict[PackageDescriptor, Package] = {}
        self._prefetched_packages: Dict[_Pkg, Promise[List[Package]]] = {}

    def _prefetch(self, package: _Pkg) -> Promise[List[Package]]:
        return get_or_put(self._prefetched_packages, package,
                          lambda: Promise.execute(self._threads, self._repo.list, package.name))

    def get_dependencies(self, package: _Pkg, version: Version) -> List[Term]:
        descriptor = PackageDescriptor(package.name, version)

        try:
            if isinstance(version, UrlVersion) and descriptor not in self.opened_packages:
                self.opened_packages[descriptor] = single_or_raise(self._repo.match(f"{package} @ {version}"))

            dependencies = self.opened_packages[descriptor] \
                .dependencies(self._env, package.extras)

            for d in dependencies:
                if not d.version_spec.specific_url():
                    self._prefetch(_Pkg.of(d))

        except (ValueError, IOError, BuildError) as e:
            raise MalformedPackageException(str(descriptor)) from e

        result: List[Term] = []

        if package.extras:  # add the package itself together with its extras
            result.append(Term(replace(package, extras=None), SpecificVersion(version)))

        for d in dependencies:
            if d.is_applicable_for(self._env, package.extras):
                if o := self._dependency_overrides.get(d.package_name):
                    result.append(Term(_Pkg.of(o), o.version_spec))
                else:
                    result.append(Term(_Pkg.of(d), d.version_spec))

        return result

    def get_versions(self, package: _Pkg) -> List[Version]:
        all_packages = self._prefetch(package).result()
        packages = [p for p in all_packages if p.is_compatible_with(self._env)]

        for package in packages:
            self.opened_packages[replace(package.descriptor, name=package.descriptor.name.lower())] = package

        return [p.version for p in packages]


@dataclass(frozen=True, eq=True)
class _Pkg:
    name: str
    extras: Optional[List[str]]

    def __str__(self):
        result = self.name
        if self.extras:
            result += "[" + ', '.join(self.extras) + "]"
        return result

    def __repr__(self):
        return self.__str__()

    def __lt__(self, other):
        return str(self) < str(other)

    def __hash__(self):
        h = hash(self.name)
        if self.extras:
            h = h * 31 + oseq_hash(self.extras)
        return h

    @classmethod
    def of(cls, d: Dependency) -> _Pkg:
        return _Pkg(d.package_name.lower(), d.extras)



import io
import mimetypes
from io import IOBase
from typing import List, Optional, Iterator, Union, IO
from uuid import uuid4

from pkm.utils.io_streams import chunks

_FORM_FIELD_SORTED_HEADERS = ["Content-Disposition", "Content-Type", "Content-Location"]


class FormField:

    def __init__(self, name: str, payload: Union[IO, str, bytes], filename: Optional[str] = None):
        self.name: str = name
        self.payload = payload
        self.filename = filename or name
        filename_header_part = f'; filename="{filename}"' if filename else ""
        self.headers = {
            "Content-Disposition": f'form-data; name="{name}"{filename_header_part}',
            "Content-Type": _guess_content_type(filename)
        }

    def set_content_type(self, content_type: str) -> "FormField":
        self.headers["Content-Type"] = content_type
        return self

    # credits: this method is adapted from urllib3.fields.RequestField.render_headers
    def render(self) -> str:
        lines = [f"{k}: {self.headers[k]}" for k in _FORM_FIELD_SORTED_HEADERS if k in self.headers]
        lines.extend(f"{k}: {v}" for k, v in self.headers.items() if v and k not in _FORM_FIELD_SORTED_HEADERS)
        lines.append("\r\n")
        return "\r\n".join(lines)


class MultipartFormDataPayload:
    def __init__(self, fields: List[FormField], chunk_size: int = io.DEFAULT_BUFFER_SIZE,
                 boundary: Optional[str] = None):
        self._fields = fields
        self._chunk_size = chunk_size
        self._boundary = boundary or uuid4().hex

    def content_type(self):
        return f"multipart/form-data; boundary={self._boundary}"

    def __iter__(self) -> Iterator[bytes]:
        boundary = f"--{self._boundary}\r\n".encode()
        newline = b"\r\n"
        chunk_size = self._chunk_size

        for field in self._fields:
            yield boundary
            yield field.render().encode()

            payload = field.payload
            if isinstance(payload, IOBase):
                yield from chunks(payload, chunk_size)
            elif isinstance(payload, str):
                yield payload.encode()
            elif isinstance(payload, bytes):
                yield payload
            else:
                raise ValueError(f'unsupported payload type for field: {field.name}: {type(field.payload)}')

            yield newline


def _guess_content_type(filename: Optional[str], default="application/octet-stream"):
    if filename:
        return mimetypes.guess_type(filename)[0] or default
    return default



from typing import TypeVar, Iterable, Callable, Dict, List, Optional, Tuple, Iterator

from pkm.utils.types import SupportHashCode

_T = TypeVar("_T")
_U = TypeVar("_U")
_K = TypeVar("_K", bound=SupportHashCode)
_SENTINAL = object()


def groupby(seq: Iterable[_T], key: Callable[[_T], _K]) -> Dict[_K, List[_T]]:
    """
    :param seq: sequence to group
    :param key: a key function, the seq items will be grouped by it
    :return:
    """
    result: Dict[_U, List[_T]] = {}
    for item in seq:
        k = key(item)
        lst = result.get(k)
        if lst is None:
            result[k] = lst = []
        lst.append(item)

    return result


def find_first(seq: Iterable[_T], match: Callable[[_T], bool], none_value: Optional[_T] = None) -> Optional[_T]:
    """
    :param seq: the sequence to look in
    :param match: predicate for items in the sequence
    :param none_value: result for no match
    :return: the first item in the sequence that match or non_value if such not found
    """
    return next((it for it in seq if match(it)), none_value)


def partition(seq: Iterable[_T], match: Callable[[_T], bool]) -> Tuple[List[_T], List[_T]]:
    """
    :param seq: the sequence to partition
    :param match: predicate for item in the sequence
    :return: two lists, the first with items that match and the second containing the rest
    """

    t, f = [], []
    for item in seq:
        if match(item):
            t.append(item)
        else:
            f.append(item)

    return t, f


def single_or_raise(seq: Iterable[_T]) -> _T:
    """
    :param seq: the iterable to access
    :return: the first element of this iterable if the number of elements in it is 1 otherwise raise `ValueError`
    """

    result = single_or(seq, _SENTINAL)
    if result is _SENTINAL:
        raise ValueError(f"expecting single element but found 0 or more than 1")

    return result


def single_or(seq: Iterable[_T], default_value: Optional[_T] = None) -> Optional[_T]:
    """
    :param seq: the iterable to access
    :param default_value: the value to return in case where iterable does not contain a single element
    :return: the first element of this iterable if the number of elements in it is 1 otherwise `default_value`
    """
    i = iter(seq)
    if (single := next(i, _SENTINAL)) is _SENTINAL:
        return default_value
    if next(i, _SENTINAL) is not _SENTINAL:
        return default_value

    return single


def distinct(it: Iterable[_T], key: Callable[[_T], _K] = lambda x: x) -> Iterator[_T]:
    """
    :param it: the iterator to run over
    :param key: key function by which to perform the comparison
    :return: iterator over `it` which only contain each item in it once
    """
    s = set()
    for item in it:
        item_key = key(item)
        if item_key not in s:
            s.add(item_key)
            yield item


def without(it: Iterable[_T], predicate: Callable[[_T], bool]) -> Iterator[_T]:
    """
    :param it: the iterator to run over
    :param predicate: predicate for items in `it`
    :return: iterator containing only items in `it` which `predicate` returned `False` for
    """
    return (v for v in it if not predicate(v))


def without_nones(it: Iterable[Optional[_T]]) -> Iterator[_T]:
    """
    :param it: the iterator to run over
    :return: iterator containing all the non-none items in `it`
    """
    return without(it, lambda x: x is None)


def first(it: Iterable[_T], default: _T) -> _T:
    """
    return the first value in the given iterable (does not have to be subscriptable)
    :param it: the iterator to run over
    :param default: the value to return if `it` if empty
    :return: the first value in `it` or `default` if `it` is empty
    """
    return next(iter(it), default)


def first_or_none(it: Iterable[_T]) -> Optional[_T]:
    """
    same as calling `first(it, None)`
    :param it: the iterator to run over
    :return: the first value in `it` or None if `it` is empty
    """
    return first(it, None)



"""
launchers are taken from:
https://bitbucket.org/vinay.sajip/simple_launcher
"""
import importlib.resources as resources
from pathlib import Path
from typing import TYPE_CHECKING
from zipfile import ZipFile

from pkm.api.distributions.distinfo import EntryPoint

if TYPE_CHECKING:
    from pkm.api.environments.environment import Environment


def build_windows_script_launcher(
        env: "Environment", target_dir: Path, file_name: str, script: str, is_gui: bool = False) -> Path:
    """
        creates windows launcher for the given `script` that is suitable to the os of the given environment.
        the launcher will be written into a file named `script_name` under the given `dir`

        :param env: the environment the created executable should use to run the script
        :param target_dir: the directory where to put the launcher in
        :param script: in case this argument is given it will be used as the script for the entrypoint,
               otherwise the script will be generated from the entrypoint object reference
        :param file_name: the base name of the executable file to create
                          (this name will be used for the file, but its extension may vary)
        :param is_gui: if True, will create a gui based launcher, defaults to False.
        :return: path to the created launcher
        """

    op_plat = env.operating_platform
    arch_suffix = '-arm' if op_plat.has_arm_cpu() else ''
    launcher_kind = 't' if is_gui else 'w'
    launcher_name = f"{launcher_kind}{arch_suffix}{op_plat.os_bits}.exe"

    launcher_data = resources.read_binary('pkm.launchers', launcher_name)

    result_file = target_dir / f"{file_name}.exe"
    with result_file.open('wb') as result_fd:
        result_fd.write(launcher_data)
        result_fd.write(f"#!{env.interpreter_path.absolute()}\r\n".encode())
        with ZipFile(result_fd) as result_zip_fd:
            result_zip_fd.writestr("__main__.py", script.encode())

    return result_file


def build_windows_entrypoint_launcher(
        env: "Environment",
        entrypoint: EntryPoint, target_dir: Path) -> Path:
    """
    creates windows launcher for the given `entrypoint` that is suitable to the os of the given environment.
    the launcher will be written into a file named `entrypoint.name` under the given `dir`

    :param env: the environment the created executable should use to run the script
    :param entrypoint: the entrypoint to create a script for
    :param target_dir: the directory where to put the launcher in
    :return: path to the created launcher
    """

    return build_windows_script_launcher(
        env, target_dir, entrypoint.name, entrypoint.ref.execution_script_snippet(),
        entrypoint.group == EntryPoint.G_CONSOLE_SCRIPTS
    )
